#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - an Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen tool

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018       Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

import argparse
import configparser
import logging
import locale
import os
import sys
import shutil

# utility diff functions
from imediff.utils import error_exit, logger, read_lines
from imediff.config import config_template, create_template
from imediff.cli import *
from imediff.tui import *

VERSION = "2.6"
PACKAGE = "imediff"

_version = (
    """\
{p} (version {v})
""".format(
        p=PACKAGE, v=VERSION
    )
    + __doc__
)

_opening = """\
===============================================================================
                               {p}                         (version {v})

The imediff command helps you to merge 2 slightly different files with an
optional base file interactively using the in-place alternating display of
the changed content on a single-pane full screen terminal user interface.

The source of line is clearly identified by the color of the line or the
identifier character at the first column.

From the console command line prompt, type:
 * "imediff" to read the tutorial,
 * "imediff -h" to get all the command line options,
 * "imediff -o output older newer" to merge 2 files, and
 * "imediff -o output yours base theirs" to merge 3 files.

For usage instructions, type "h" and "H" in the interactive screen.

 * Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
 * Copyright (C) 2018       Osamu Aoki <osamu@debian.org>

License: GPL 2.0+
===============================================================================
""".format(
    p=PACKAGE, v=VERSION
)


def initialize_args():
    """
    Parse command line options and arguments

    Input:
        commandline

    Return:
        args    argument values
    """
    pa = argparse.ArgumentParser(
        description="Interactive Merge Editor for 2 or 3 different files",
        epilog='Start "imediff" without any arguments to see the tutorial.',
    )
    group = pa.add_mutually_exclusive_group()
    group.add_argument(
        "-a", action="store_true", help="Start with all chunks to use file_a"
    )
    group.add_argument(
        "-b", action="store_true", help="Start with all chunks to use file_b"
    )
    group.add_argument(
        "-c",
        action="store_true",
        help="Start with all chunks to use file_c (only for diff3)",
    )
    group.add_argument(
        "-d", action="store_true", help="Start with all chunks to use diff"
    )
    group.add_argument(
        "-u", action="store_true", help=argparse.SUPPRESS
    )  # Start with unresolved diff
    group.add_argument(
        "-f", action="store_true", help="Start with all chunks to use wdiff"
    )
    group.add_argument(
        "-g", action="store_true", help="Start with good merge mode (only for diff3)"
    )
    group.add_argument(
        "--isjunk",
        action="store_true",
        help="Force isjunk to None instead of the default list",
    )
    pa.add_argument(
        "--linerule",
        "-l",
        action="store",
        default=2,
        help="Line alignment matching rule (0,1,2,3,10,11,12,13)",
    )
    pa.add_argument("--mode", "-m", action="store_true", help="Display mode column")
    pa.add_argument("--mono", action="store_true", help="Force monochrome display")
    pa.add_argument(
        "--sloppy", action="store_true", help="Allow one to save unresolved contents"
    )
    pa.add_argument(
        "--version", "-V", action="store_true", help="Show version and license"
    )
    pa.add_argument(
        "--output",
        "-o",
        help="\
Write output to the given file.  If this is missing, use STDERR",
    )
    pa.add_argument(
        "--conf",
        "-C",
        default="~/.imediff",
        help='\
Specify configuration file to use.  (default="~/.imediff")',
    )
    # hidden option for debug: non-interactive diff/merge operation
    pa.add_argument(
        "--non-interactive", "-n", action="store_true", help="execution without curses"
    )
    pa.add_argument("--macro", "-M", default=":", help="set MACRO string")
    pa.add_argument(
        "--template",
        "-t",
        action="store_true",
        help='Create a template configuration file "~/.imediff"',
    )
    pa.add_argument(
        "--debug", "-D", action="store_true", help='Generate debug log in "imediff.log"'
    )
    pa.add_argument("file_a", nargs="?", help="file for OLDER(diff2), YOURS(diff3)")
    pa.add_argument("file_b", nargs="?", help="file for NEWER(diff2), BASE(diff3)")
    pa.add_argument(
        "file_c",
        nargs="?",
        help="file for ------------, THEIRS(diff3) (only for diff3)",
    )
    args = pa.parse_args()
    args.macro_buffer = args.macro
    if args.file_c is not None:
        args.diff_mode = 3
    elif args.file_b is not None:
        args.diff_mode = 2
    else:
        args.diff_mode = 0
    if args.a:
        args.default_mode = "a"
    elif args.b:
        args.default_mode = "b"
    elif args.c and args.diff_mode == 2:
        args.default_mode = "d"  # backward compatibility -c
    elif args.c and args.diff_mode == 3:
        args.default_mode = "c"
    elif args.d or args.u:
        args.default_mode = "d"  # hidden backward compatibility -u
    elif args.f:
        args.default_mode = "f"
    else:  # default=None or "g"
        if args.diff_mode == 2:
            args.default_mode = "d"  # default diff2
        else:
            args.default_mode = "g"  # default diff3
    return args


def initialize_confs(conf):
    """Process configuration file"""
    config_file = os.path.expanduser(conf)
    # Allow inline comment with #
    confs_i = configparser.ConfigParser(inline_comment_prefixes=("#"))
    confs_i.read_string(config_template)
    confs_f = configparser.ConfigParser(inline_comment_prefixes=("#"))
    if os.path.exists(config_file):
        confs_f.read(config_file)
        if (
            "version" in confs_f["config"].keys()
            and confs_f["config"]["version"] == confs_i["config"]["version"]
        ):
            confs = confs_f
        else:
            error_exit(
                '''\
Error in {0}: version mismatch
        the current version:  {1}
        the required version: {2}

Rename {0} to {0}.bkup and make the new {0} by
editing the template obtained by "imediff -t"'''.format(
                    conf, confs_f["config"]["version"], confs_i["config"]["version"]
                )
            )
    else:
        confs = confs_i
    return confs


##############################################################################
def main():
    """
    Entry point for imediff command

    Exit value
        0       program exits normally after saving data
        1       program quits without saving
        2       program terminates after an internal error
    """

    # preparation and arguments
    locale.setlocale(locale.LC_ALL, "")
    args = initialize_args()
    if args.template:
        create_template(args.conf)
        sys.exit(0)

    # logging
    logger.setLevel(logging.DEBUG)
    if args.debug:
        # create file handler which logs even debug messages
        fh = logging.FileHandler("imediff.log")
    else:
        # create file handler which doesn't log
        fh = logging.NullHandler()
    fh.setLevel(logging.DEBUG)
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    # create formatter and add it to the handlers
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)

    # configuration
    confs = initialize_confs(args.conf)
    editor = "editor"
    if "EDITOR" in os.environ:
        editor = os.environ["EDITOR"]
    if "editor" in confs["config"].keys():
        editor = confs["config"]["editor"]
    args.edit_cmd = shutil.which(editor)
    if args.edit_cmd is None:
        args.edit_cmd = "/usr/bin/editor"  # safe fall back
    logger.debug("external editor {} found as {}".format(editor, args.edit_cmd))

    # normalize and process non-standard situation
    if args.version:
        print(_version)
        sys.exit(0)
    if args.diff_mode == 0 and args.non_interactive:
        tutorial = True
    if args.diff_mode == 0:
        args.diff_mode = 3  # Fake input
        list_a = (_opening + '\n    Type "q" to quit this tutorial.').splitlines(
            keepends=True
        )
        list_b = [""]
        list_c = list_a
        confs["config"]["confirm_quit"] = "False"
        confs["config"]["confirm_exit"] = "False"
        tutorial = True
    elif args.diff_mode == 2:
        # diff2
        list_a = read_lines(args.file_a)
        list_b = read_lines(args.file_b)
        list_c = None
        tutorial = False
    elif args.diff_mode == 3:
        list_a = read_lines(args.file_a)
        list_b = read_lines(args.file_b)
        list_c = read_lines(args.file_c)
        tutorial = False
    else:
        error_exit("imediff normally takes 2 or 3 files")

    # call main routine
    if not args.non_interactive:
        display_instance = TextPad(list_a, list_b, list_c, args, confs)
        display_instance.command_loop(tutorial=tutorial)
        del display_instance
    elif tutorial:
        print(_opening)
    else:
        text_instance = TextData(list_a, list_b, list_c, args, confs)
        text_instance.command_loop()
        del text_instance
    sys.exit(0)



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - an Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen tool

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018-2021  Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

import curses

from imediff.utils import _, console_width, error_exit, logger, write_file
from imediff.config import cc
from imediff.cli import *

# Format stings
_helptext2 = _(
    """\
    Interactive Merge Editor to merge 2 DIFFerent files

imediff merges 2 different input files using 5 modes:
 * mode a: display {file_a} (older) in {color_a}
 * mode b: display {file_b} (newer) in {color_b}
 * mode d: display diff2(a,b) by line       in {color_d}
 * mode e: display editor result buffer     in {color_e}
 * mode f: display wdiff2(a,b) by character in {color_f}

key commands          induced actions
{w:c},{x:c}                   write and exit
{q:c}                     quit without saving
{a:c}/{b:c}/{d:c}/{e:c}/{f:c}             set a chunk to a/b/d/e/f mode
1/2/4/5/6             set a chunk to a/b/d/e/f mode (alternative)
{A:c}/{B:c}/{D:c}/{E:c}/{F:c}             set all chunks to a/b/d/e/f mode
enter                 toggle mode of a chunk
{m:c}                     modify a chunk with editor: {edit_cmd}
{M:c}                     remove a editor result buffer
arrows/pgdn,{j:c}/pgup,{k:c}  move scope of the display
space,{n:c} /backspace,{p:c}  select the next/previous chunk
tab,{N:c}   /shift-tab,{P:c}  select the next/previous diff chunk
home,{t:c}  /end,{z:c}        select the first/last chunk
{T:c}       /{Z:c}            select the first/last diff chunk
?,{s:c}                   show the state of the merge
{h:c}                     show this help
{H:c}                     show tutorial"""
)

_helptext3 = _(
    """\
    Interactive Merge Editor to merge 3 DIFFerent files

imediff merges 3 different input files using 7 modes:
 * mode a: display {file_a} (yours) in {color_a}
 * mode b: display {file_b} (base)  in {color_b}
 * mode c: display {file_c} (their) in {color_c}
 * mode d: display diff3(a,b,c) by line       in {color_d}
 * mode e: display editor result buffer       in {color_e}
 * mode f: display wdiff3(a,b,c) by character in {color_f}
 * mode g: set good mode from (a,b,c,e,g) if merged cleanly, or
           set mode (df) in case of conflicts

key commands          induced actions
{w:c},{x:c}                   write and exit
{q:c}                     quit without saving
{a:c}/{b:c}/{c:c}/{d:c}/{e:c}/{f:c}/{g:c}         set a chunk to a/b/c/d/e/f/g mode
1/2/3/4/5/6/7         set a chunk to a/b/c/d/e/f/g mode (alternative)
{A:c}/{B:c}/{C:c}/{D:c}/{E:c}/{F:c}/{G:c}         set all chunks to a/b/c/d/e/f/g mode
enter                 toggle mode of a chunk
{m:c}                     modify a chunk with editor: {edit_cmd}
{M:c}                     remove a editor result buffer
arrows/pgdn,{j:c}/pgup,{k:c}  move scope of the display
space,{n:c} /backspace,{p:c}  select the next/previous chunk
tab,{N:c}   /shift-tab,{P:c}  select the next/previous diff chunk
home,{t:c}  /end,{z:c}        select the first/last chunk
{T:c}       /{Z:c}            select the first/last diff chunk
?,{s:c}                   show the state of the merge
{h:c}                     show this help
{H:c}                     show tutorial"""
)

_stattext0 = _(
    """\
Chunk index   = (all files are identical)
Line index    ={row: 5d} (total ={conth: 5d})
Column offset ={col: 5d}"""
)

_stattext1 = _(
    """\
Chunk index   ={active: 5d} (total ={total: 5d}, unresolved ={unresolved: 5d})
Line index    ={row: 5d} (total ={conth: 5d})
Column offset ={col: 5d}"""
)

# Keep this under 74 char/line for better looks
_nonclean = _(
    """\
This requirement of the clean merge for 'save and exit' can be disabled by
specifying the "--sloppy" option to the imediff command.  Alternatively,
you can effectively evade this requirement by pressing "m" on all
non-clean merges to make them as the manually edited text data."""
)

# Keep this under 74 char/line for better looks
# I need this hack to avoid translation of tutorial for now. XXX FIXME XXX
nonclean = """\
This requirement of the clean merge for 'save and exit' can be disabled by
specifying the "--sloppy" option to the imediff command.  Alternatively,
you can effectively evade this requirement by pressing "m" on all
non-clean merges to make them as the manually edited text data."""


# Keep this under 76 char/line to fit this in the 80 char terminal
_tutorial = """\
Quick start:
    In this tutorial screen, use pgdn/pgup/arrows keys to read it, or
    type other keys to return to the main imediff screen.

---------------------------------------------------------------------------
    Tutorial for imediff (Interactive Merge Editor)
                        Copyright (C) 2021 Osamu Aoki <osamu@debian.org>
---------------------------------------------------------------------------

The imediff command helps you to merge 2 slightly different files with an
optional base file interactively using the in-place alternating display of
the changed content on a single-pane full screen terminal user interface.

The source of line is clearly identified by the color of the line or the
identifier character at the first column.

The advantage of this user interface is the minimal movement of the line of
sight for the user.  Other great tools such as vimdiff, xxdiff, meld and
kdiff3 require you to look at different points of display to find the exact
position of changes.  This makes imediff the most stress-free tool.  (I
realized this when I first used the original imediff2 program by Jarno
Elonen <elonen@iki.fi>.  Please note that the command name is changed from
imediff2 to imediff now.)

Other great tools for merge such as "diff3 -m ..." and "git merge ..."
operate only on the difference by line.  So even for the non-overlapping
changes, they yield the merge conflict if changes happen on the same line.

The automatic merge logic of the imediff command operates not only on the
difference by line but on the difference by character.  This is another
great feature of the imediff command. So for the non-overlapping changes, it
always yields the clean merge (mode "g").

Merge with 2 files
==================

Let's try to merge 2 almost identical files, "file_a" and "file_b", into an
output file, "file_o".  You can do this with the following.

    $ imediff -o file_o file_a file_b

This starts a full screen display of the content of the intended output
"file_o".  Initially all the different lines in "file_a" and "file_b" are
grouped in chunks and displayed in mode "d" which combines the corresponding
"file_a" and "file_b" content by line.

You can move the focused chunk to the next chunk by pressing "Space", or "n"
keys.  You can move the focused chunk to the previous chunk by pressing
"Back Space", or "p" keys.

You can set the display mode of the focused chunk with the corresponding
single key command.  Pressing "a" displays the "file_a" content.  Pressing
"b" displays the "file_b" content.

By alternating "a" and "b" keys, you can see the difference in place which
is easy on you with the constant line of sight.  (This is the same great
feature inherited from the original imediff2 program.)

You can display both the "file_a" content and the "file_b" content with 2
key commands.  Pressing "d" displays 2 blocks of lines organized somewhat
like "diff -u" (mode "d").  Pressing "f" displays intermixed 1 block of
lines organized somewhat like "wdiff" (mode "f").

Pressing "m" starts an editor to edit the focused chunk from any modes to
create a manually merged content.  Upon exiting the editor, its result is
kept in the editor result buffer.  Even after pressing "a", "b", "d", or
"f", the content of the editor result buffer can be recalled and displayed
by pressing "e".

Pressing "M" in mode "e" removes the editor result buffer.

When you press one of the upper case "A", "B", "D", "E", "F", this sets
all chunks to the corresponding lower case mode.

All parts of imediff data normally need to select "a", "b", or "e" (excluding
"d" and "f") before writing the merge result unless "--sloppy" is specified
before writing the result.  Type "w" or "x" to write the displayed content to
"file_o" and exit the imediff program.

{}

Although the imediff program is practically WYSIWYG, there is one notable
exception.  For the deleted content in mode "a" or "b", the imediff program
displays "???" in reverse mode as a placeholder.  This "???" is not included
in the output file.

Merge with 3 files
==================

Let's try to merge 2 almost identical files, "file_a" and "file_c", both of
which are based on the file, "file_b", into an output file, "file_o".  You
can do this with the following.

    $ imediff -o file_o file_a file_b file_c

This starts a full screen display of the intended merged output content
"file_o".  Unlike "Merge with 2 files", the existence of the common base
file "file_b" allows the imediff program to chose desirable chunks
automatically from the corresponding "file_a" or "file_c" contents like
"diff3 -m" or "git merge".

Actually, this imediff program does more.  This imediff program merges not
only with comparison by the line like other tools but also comparison by the
character.  This allows the clean automatic merge even when changes happen
on different positions of the same line of "file_a" and "file_c" derived
from "file_b".

Only really unresolved chunks are displayed in mode "d".  You can move the
focused chunk to the next unresolved chunk by pressing "Tab", or "N" keys.
Pressing "m" starts an editor to create a manually merged content.  You can
move the focused chunk to the previous unresolved chunk by pressing
"Shift-Tab", or "P" keys.

The key binding for "Merge with 3 files" is almost the same as that for
"Merge with 2 files".  There are 2 notable extensions.  Pressing "c"
displays the "file_c" content.  Pressing "g" causes automatic merge efforts
on a chunk for 3 files in the following order:
 * If the editor result buffer has content, mode is set to "e".
 * If a chunk is resolved cleanly, mode is set to "a", "c", or "g".
   This overrides previous manual settings such as "a", "b", or "c".
 * If a chunk isn't resolved cleanly, mode is left as mode "d" or "f".

By alternating "a" and "c" keys, you can see the difference in place.

All parts of imediff data normally need to select "a", "b", "c", "g" or "e"
(excluding "d" and "f") before writing the merge result unless "--sloppy" is
specified before writing the result.

The rests are mostly the same as "Merge with 2 files".

Terminal
========

The imediff program is compatible with any terminal window sizes.  It
supports both monochrome and color terminals.  For comfortable user
experience, terminal width of 80 characters/line or more and terminal height
of 24 lines or more are desirable.

When mode column is enabled by "-m" option or monochrome terminal is used,
the first column is used to indicate the match type of each un-selectable
matched section or display mode of each selectable unmatched chunk.
 * "=" for a un-selectable section means unchanged source files: a==b==c.
   (Under the color terminal, this is displayed in white/normal.)
 * "#" for a un-selectable section means matched and changed files: a==c and
   a!=b.  (Under the color terminal, this is displayed in white/bold.)
 * "a", "b", "c", "d", "e", "f", and "g" for selectable unmatched chunk are
   the display mode of each chunk.  (Under the color terminal, these are
   displayed in different colors.)

Customization
=============

The imediff program can customize its key binding and its color setting with
the "~/.imediff" file in the ini file format.  You can create its template
file by the "imediff -t" command.

You can disable an existing "~/.imediff" file without renaming it by
specifying a non-existing file, e.g., "BOGUS" as "imediff -C BOGUS ...".
The internal default values of the imediff program are used.

Note
====

Some keys are aliased for "Merge with 2 files" for your convenience:
 * "c" works as "d"
 * "g" works as "e"

The "diff3 -m" has an odd feature for the merge when "file_a" and "file_c"
undergo identical changes from "file_b".  This imediff program results in a
more intuitive merge result.""".format(
    nonclean
)


class TextPad(TextData):  # TUI data
    """Curses class to handle diff data for 2 or 3 lines"""

    # persistent variables: self.*
    #     stdscr: display object
    #     winh, winw: Window height and width from stdscr.getmaxyx()
    #     conth, contw: Content height and width of textpad
    #     row, col: window top left position in textpad
    #     active: index for self.actives
    # non-persistent running variables
    #     i: index as self.opcodes[i]
    #     j: index as self.actives[j]
    def __init__(self, list_a, list_b, list_c, args, confs):
        # Init from super class "TextData"
        super().__init__(list_a, list_b, list_c, args, confs)
        # Init from commandline/configuration parameters
        self.mode = args.mode
        self.mono = args.mono
        if self.diff_mode == 2:
            self.color = confs["color_diff2"]
        else:
            self.color = confs["color_diff3"]
        return

    def command_loop(self, tutorial=False):  # for curses TUI (wrapper)
        self.tutorial = tutorial
        curses.wrapper(self.gui_loop)
        return

    def gui_loop(self, stdscr):  # for curses TUI (core)
        # initialize
        self.stdscr = stdscr
        color = self.color  # shorthand
        self.winh, self.winw = self.stdscr.getmaxyx()  # window size
        curses.start_color()
        self.stdscr.clear()
        self.stdscr.refresh()
        # set color pair_number as (pair_number, fg, bg)
        curses.init_pair(1, cc[color["color_a"]], cc["BLACK"])
        curses.init_pair(2, cc[color["color_b"]], cc["BLACK"])
        curses.init_pair(3, cc[color["color_c"]], cc["BLACK"])
        curses.init_pair(4, cc[color["color_d"]], cc["BLACK"])
        curses.init_pair(5, cc[color["color_e"]], cc["BLACK"])
        curses.init_pair(6, cc[color["color_f"]], cc["BLACK"])
        #
        # +6: active cc
        self.active_color = 6
        curses.init_pair(7, cc["WHITE"], cc[color["color_a"]])
        curses.init_pair(8, cc["WHITE"], cc[color["color_b"]])
        curses.init_pair(9, cc["WHITE"], cc[color["color_c"]])
        curses.init_pair(10, cc["WHITE"], cc[color["color_d"]])
        curses.init_pair(11, cc["WHITE"], cc[color["color_e"]])
        curses.init_pair(12, cc["WHITE"], cc[color["color_f"]])
        #
        # +12: deleted cc
        self.deleted_color = 12
        curses.init_pair(13, cc[color["color_a"]], cc["WHITE"])
        curses.init_pair(14, cc[color["color_b"]], cc["WHITE"])
        curses.init_pair(15, cc[color["color_c"]], cc["WHITE"])
        curses.init_pair(16, cc[color["color_d"]], cc["WHITE"])
        curses.init_pair(17, cc[color["color_e"]], cc["WHITE"])
        curses.init_pair(18, cc[color["color_f"]], cc["WHITE"])
        #
        # +6+12
        curses.init_pair(19, cc["BLACK"], cc[color["color_a"]])
        curses.init_pair(20, cc["BLACK"], cc[color["color_b"]])
        curses.init_pair(21, cc["BLACK"], cc[color["color_c"]])
        curses.init_pair(22, cc["BLACK"], cc[color["color_d"]])
        curses.init_pair(23, cc["BLACK"], cc[color["color_e"]])
        curses.init_pair(24, cc["BLACK"], cc[color["color_f"]])
        #
        if curses.has_colors() == False:
            self.mono = True
        if self.mono:
            self.mode = True
            self.color_a = "WHITE"
            self.color_b = "WHITE"
            self.color_c = "WHITE"
            self.color_d = "WHITE"
            self.color_e = "WHITE"
            self.color_f = "WHITE"
        else:
            if self.diff_mode == 2:
                self.color_a = color["color_a"]
                self.color_b = color["color_b"]
                # self.color_c = color['color_c'] # never used
                self.color_d = color["color_d"]
                self.color_e = color["color_e"]
                self.color_f = color["color_f"]
            else:
                self.color_a = color["color_a"]
                self.color_b = color["color_b"]
                self.color_c = color["color_c"]
                self.color_d = color["color_d"]
                self.color_e = color["color_e"]
                self.color_f = color["color_f"]
        # display parameters
        self.col = 0  # the column coordinate of textpad (left most=0)
        self.row = 0  # the row coordinate of textpad    (top most=0)
        self.update_textpad = True  # update textpad content
        while True:
            if self.active is not None:
                logger.debug(
                    "command loop: active = {} active_index = {} row = {} col = {}".format(
                        self.active, self.actives[self.active], self.row, self.col
                    )
                )
            else:
                logger.debug(
                    "command loop: active = ***None*** row = {} col = {}".format(
                        self.row, self.col
                    )
                )
            curses.curs_set(0)
            if self.update_textpad:
                self.new_textpad()
            # clear to remove garbage outside of textpad
            self.winh, self.winw = self.stdscr.getmaxyx()
            self.adjust_window()
            for icol in range(self.contw - self.col, self.winw):
                self.stdscr.vline(0, icol, " ", self.winh)
            ##self.stdscr.vline(0, self.contw - self.col, '@', self.winh)
            ##self.stdscr.vline(0, self.winw-1, '*', self.winh)
            # clear rows downward to remove garbage characters
            for irow in range(self.conth - self.row, self.winh):
                self.stdscr.hline(irow, 0, " ", self.winw)
            ##if (self.conth - self.row) <= self.winh -1 and (self.conth - self.row) >= 0:
            ##    self.stdscr.hline(self.conth - self.row , 0, '@', self.winw)
            if self.update_textpad or self.update_active:
                self.highlight()
            self.textpad.refresh(self.row, self.col, 0, 0, self.winh - 1, self.winw - 1)
            if self.active is not None:
                row = self.get_row(self.actives[self.active]) - self.row
                if row >= 0 and row < self.winh:
                    self.stdscr.move(row, 0)
                    curses.curs_set(1)
                else:
                    curses.curs_set(0)
            self.stdscr.refresh()
            # reset flags
            self.update_textpad = False
            self.update_active = False
            if self.tutorial:
                c = ord("H")
                self.tutorial = False
            else:
                c = self.getch_translated()
            ch = chr(c)
            if ch == "w" or ch == "x" or c == curses.KEY_EXIT or c == curses.KEY_SAVE:
                if self.sloppy or (
                    not self.sloppy and self.get_unresolved_count() == 0
                ):
                    if not self.confirm_exit or self.popup(
                        _("Do you 'save and exit'? (Press '{y:c}' to exit)").format(
                            y=self.rkc["y"]
                        )
                    ):
                        output = self.get_output()
                        write_file(self.file_o, output)
                        break
                else:
                    self.popup(
                        _(
                            "Can't 'save and exit' due to the non-clean merge. (Press '{y:c}' to continue)"
                            + "\n\n"
                            + _nonclean
                        ).format(y=self.rkc["y"])
                    )
            elif ch == "q":
                if not self.confirm_exit or self.popup(
                    _("Do you 'quit without saving'? (Press '{y:c}' to quit)").format(
                        y=self.rkc["y"]
                    )
                ):
                    self.opcodes = []
                    error_exit("Quit without saving by the user request\n")
            elif ch == "h" or c == curses.KEY_HELP:
                # Show help screen
                self.popup(self.helptext())
            elif ch == "H":
                # Show tutorial screen
                self.popup(_tutorial)
            elif ch == "s" or ch == "?":
                # Show location
                if len(self.actives) == 0:
                    self.popup(
                        _stattext0.format(row=self.row, conth=self.conth, col=self.col)
                    )
                else:
                    self.popup(
                        _stattext1.format(
                            active=self.active,
                            total=len(self.actives),
                            unresolved=self.get_unresolved_count(),
                            row=self.row,
                            conth=self.conth,
                            col=self.col,
                        )
                    )
            # Moves in document
            elif c == curses.KEY_SR or c == curses.KEY_UP or ch == "k":
                self.row -= 1
            elif c == curses.KEY_SF or c == curses.KEY_DOWN or ch == "j":
                self.row += 1
            elif c == curses.KEY_LEFT:
                self.col -= 8
            elif c == curses.KEY_RIGHT:
                self.col += 8
            elif c == curses.KEY_PPAGE:
                self.row -= self.winh
            elif c == curses.KEY_NPAGE:
                self.row += self.winh
            # Terminal resize signal
            elif c == curses.KEY_RESIZE:
                self.winh, self.winw = self.stdscr.getmaxyx()
            else:
                pass
            # Following key-command updates TextPad
            if self.active is not None:
                # get active chunk
                # Explicitly select chunk mode
                if ch in "abdef":
                    self.set_mode(self.actives[self.active], ch)
                elif ch in "12456":
                    self.set_mode(
                        self.actives[self.active], chr(ord(ch) - ord("1") + ord("a"))
                    )
                elif ch in "ABDEF":
                    self.set_all_mode(ch.lower())
                elif ch in "cg" and self.diff_mode == 3:
                    self.set_mode(self.actives[self.active], ch)
                elif ch in "37" and self.diff_mode == 3:
                    self.set_mode(
                        self.actives[self.active], chr(ord(ch) - ord("1") + ord("a"))
                    )
                elif ch in "CG" and self.diff_mode == 3:
                    self.set_all_mode(ch.lower())
                elif c == 10 or c == curses.KEY_COMMAND:
                    mode = self.get_mode(self.actives[self.active])
                    if mode == "a":
                        self.set_mode(self.actives[self.active], "b")
                    elif mode == "b" and self.diff_mode == 2:
                        self.set_mode(self.actives[self.active], "d")
                    elif mode == "b" and self.diff_mode == 3:
                        self.set_mode(self.actives[self.active], "c")
                    elif mode == "c":
                        self.set_mode(self.actives[self.active], "d")
                    elif (
                        mode == "d"
                        and self.get_bf(self.actives[self.active]) is not None
                    ):
                        self.set_mode(self.actives[self.active], "e")
                    elif mode == "d":
                        self.set_mode(self.actives[self.active], "f")
                    elif mode == "e":
                        self.set_mode(self.actives[self.active], "f")
                    else:  # f
                        self.set_mode(self.actives[self.active], "a")
                elif ch == "m":
                    self.editor(self.actives[self.active])
                elif ch == "M" and mode == "e":
                    self.del_editor(self.actives[self.active])
                elif ch == "n" or c == curses.KEY_NEXT or ch == " ":
                    self.active_next()
                elif ch == "p" or c == curses.KEY_PREVIOUS or c == curses.KEY_BACKSPACE:
                    self.active_prev()
                elif ch == "t" or c == curses.KEY_HOME:
                    self.active_home()
                elif ch == "z" or c == curses.KEY_END:
                    self.active_end()
                elif ch == "N" or ch == "\t":
                    self.diff_next()
                elif ch == "P" or c == curses.KEY_BTAB:
                    self.diff_prev()
                elif ch == "T":
                    self.diff_home()
                elif ch == "Z":
                    self.diff_end()
                else:
                    pass
            logger.debug("command-loop")
        return

    def new_textpad(self):
        """Create new curses textpad"""
        # pre-scan content to get big enough textpad size
        conth = 0  # content height
        contw = 0  # content width
        for i in range(len(self.opcodes)):
            self.set_row(i, conth)  # record textpad row position in chunk
            tag = self.get_tag(i)
            content = self.get_content(i)  # list()
            conth += len(content)
            if tag == "E" or tag == "e":
                pass
            else:
                if len(content) == 0:
                    conth += 1
            for line in content:
                contw = max(contw, console_width(line))
        if self.mode:  # Add mode column
            contw += 2  # for the tag indicator + ' '
        self.conth = conth
        self.contw = contw
        # actual textpad size slightly bigger for safety margin
        self.textpad = curses.newpad(conth + 1, max(80, contw + 1))
        for i in range(len(self.opcodes)):
            self.textpad_addstr(i, False)
        if self.active is not None:
            logger.debug(
                "gui init: active={} chunk_index={} row={} col={} conth={} contw={}".format(
                    self.active,
                    self.actives[self.active],
                    self.row,
                    self.col,
                    self.conth,
                    self.contw,
                )
            )
        return

    def textpad_addstr(self, i, selected=False):
        tag = self.get_tag(i)
        mode = self.get_mode(i)
        if tag == "E":  # Same a = b = c
            decor = curses.A_DIM
            color_pair = 0
            prefix = "= "
        elif tag == "e":  # Same a = c
            decor = curses.A_BOLD
            color_pair = 0
            prefix = "# "
        elif mode == "a":  # diff2 OLD  /diff3: YOURS NEW
            decor = curses.A_BOLD
            color_pair = 1
            prefix = mode + " "
        elif mode == "b":  # diff2 NEW  /diff3: common OLD
            decor = curses.A_BOLD
            color_pair = 2
            prefix = mode + " "
        elif mode == "c":  # diff2 ---  /diff3: THEIRS NEW
            decor = curses.A_BOLD
            color_pair = 3
            prefix = mode + " "
        elif mode == "d":  # diff
            decor = curses.A_BOLD
            color_pair = 4
            prefix = mode + " "
        elif mode == "e":  # edit buffer
            decor = curses.A_BOLD
            color_pair = 5
            prefix = mode + " "
        elif mode == "f":  # wdiff
            decor = curses.A_BOLD
            color_pair = 6
            prefix = mode + " "
        else:  # 'g': # wdiff, cleanly merged
            decor = curses.A_BOLD
            decor |= curses.A_REVERSE
            color_pair = 6
            prefix = mode + " "
        row = self.get_row(i)
        content = self.get_content(i)  # list()
        # Decorative "???" for deleted lines only for display
        if len(content) == 0 and tag not in "Ee":
            content = ["???"]  # override []
            if self.mono:
                decor |= curses.A_REVERSE
            else:
                color_pair += self.deleted_color
        if selected:
            color_pair += self.active_color
            decor |= curses.A_REVERSE
        for line in content:
            # logger.debug("textpad.addstr, >>>  row={} line={} decor={} color_pair={}".format(row, line[:-1], decor , color_pair))
            if self.mono:
                self.textpad.addstr(row, 0, prefix + line, decor)
            elif self.mode:
                self.textpad.addstr(
                    row, 0, prefix + line, decor | curses.color_pair(color_pair)
                )
            else:
                self.textpad.addstr(row, 0, line, decor | curses.color_pair(color_pair))
            row += 1
        return

    def adjust_window(self):
        """Clamp window scope to have cursor within window and content"""
        # row,   col   -> index:  first column/line = 0
        # winh,  winw  -> number: first column/line = 1
        # conth, contw -> number: first column/line = 1
        context_length = 10
        self.winh, self.winw = self.stdscr.getmaxyx()
        if self.update_active and self.active is not None:
            selected_row = self.get_row(self.actives[self.active])
            if self.row >= selected_row - context_length:
                self.row = selected_row - context_length
            if self.row <= selected_row - self.winh + context_length:
                self.row = selected_row - self.winh + context_length
        # clamp cursor col/row within data
        if self.row >= self.conth - self.winh:
            self.row = self.conth - self.winh
        if self.row < 0:
            self.row = 0
        if self.col >= self.contw - 1:
            self.col = self.contw - 1
        if self.col < 0:
            self.col = 0
        logger.debug(
            "adjust_window: conth={} contw={} winh={} winw={} row={}, col={}".format(
                self.conth, self.contw, self.winh, self.winw, self.row, self.col
            )
        )
        return

    def highlight(self):
        """Update textpad by repainting active highlight"""
        if self.active is None:
            # No diff --> No highlight action
            # Avoid out of range for self.active, self.active_old
            # This makes highlight more robust and easy to use
            pass
        else:
            if self.active != self.active_old and self.active_old is not None:
                # Repaint old selection without highlighting
                self.textpad_addstr(self.actives[self.active_old], selected=False)
            self.textpad_addstr(self.actives[self.active], selected=True)
        return

    def set_mode(self, i, new_mode):
        super().set_mode(i, new_mode)
        self.update_textpad = True
        return

    def getch_translated(self):
        """Macro parsing instead of curses getch"""
        if len(self.macro):
            c = ord(self.macro[:1])
            c = self.c_translated(c)
            if c == ord(":"):
                try:
                    c = self.stdscr.getch()
                except:
                    c = ord("q")  # quit w/o saving for ^C
                c = self.c_translated(c)
            else:
                self.macro = self.macro[1:]
        else:
            c = 0  # End of MACRO
        return c

    def popup(self, text):
        self.winh, self.winw = self.stdscr.getmaxyx()
        popupw = 0
        popuph = 0
        for line in text.split("\n"):
            popupw = max(popupw, console_width(line))
            popuph += 1
        popuph = popuph + 2  # top/bottom border
        popupw = popupw + 4  # left/right (border + space)
        popuppad = curses.newpad(popuph, popupw)
        for i, line in enumerate(text.split("\n")):
            popuppad.addstr(1 + i, 2, line, curses.A_BOLD)
        popuppad.border()
        poprow = 0
        popcol = 0
        curses.curs_set(0)
        while True:
            self.winh, self.winw = self.stdscr.getmaxyx()
            popwinh = min(popuph, self.winh)
            popwinw = min(popupw, self.winw)
            self.textpad.refresh(self.row, self.col, 0, 0, self.winh - 1, self.winw - 1)
            # logger.debug("popup before >>>  poprow={} popcol={} popupw={} popwinw={}".format(poprow, popcol, popupw, popwinw))
            if poprow <= 0:
                poprow = 0
            if poprow >= popuph - popwinh:
                poprow = popuph - popwinh
            if popcol <= 0:
                popcol = 0
            if popcol >= popupw - popwinw:
                popcol = popupw - popwinw
            popuppad.refresh(
                poprow,
                popcol,
                max((self.winh - popwinh) // 2, 0),
                max((self.winw - popwinw) // 2, 0),
                min((self.winh + popwinh + 1) // 2 - 1, self.winh - 1),
                min((self.winw + popwinw + 1) // 2 - 1, self.winw - 1),
            )
            self.stdscr.refresh()
            c = self.getch_translated()
            ch = chr(c)
            # Moves in document
            if c == curses.KEY_SR or c == curses.KEY_UP or ch == "k":
                poprow -= 1
            elif c == curses.KEY_SF or c == curses.KEY_DOWN or ch == "j":
                poprow += 1
            elif c == curses.KEY_LEFT:
                popcol -= 1
            elif c == curses.KEY_RIGHT:
                popcol += 1
            elif c == curses.KEY_PPAGE:
                poprow -= popwinh
            elif c == curses.KEY_NPAGE:
                poprow += popwinh
            # Terminal resize signal
            elif c == curses.KEY_RESIZE:
                self.winh, self.winw = self.stdscr.getmaxyx()
            elif c == ord("y") or c == ord("y") - 32:
                result = True
                break
            else:
                result = False
                break
            self.stdscr.refresh()
        return result

    def editor(self, i):
        # logger.debug("Before invoking editor")
        self.stdscr.keypad(0)
        curses.savetty()
        curses.echo()
        curses.nocbreak()
        curses.endwin()
        super().editor(i)
        curses.cbreak()
        curses.noecho()
        curses.resetty()
        self.stdscr.keypad(True)  # keys processed by curses (again)
        self.stdscr.clear()
        self.stdscr.refresh()
        # logger.debug("After invoking editor")
        return

    def helptext(self):
        if self.diff_mode == 2:
            text = _helptext2.format(
                file_a=self.file_a,
                color_a=self.color_a,
                file_b=self.file_b,
                color_b=self.color_b,
                color_d=self.color_d,
                color_e=self.color_e,
                color_f=self.color_f,
                w=self.rkc["w"],
                x=self.rkc["x"],
                q=self.rkc["q"],
                a=self.rkc["a"],
                b=self.rkc["b"],
                d=self.rkc["d"],
                e=self.rkc["e"],
                f=self.rkc["f"],
                A=self.rkc["a"] - 32,
                B=self.rkc["b"] - 32,
                D=self.rkc["d"] - 32,
                E=self.rkc["e"] - 32,
                F=self.rkc["f"] - 32,
                m=self.rkc["m"],
                edit_cmd=self.edit_cmd,
                M=self.rkc["m"] - 32,
                j=self.rkc["j"],
                k=self.rkc["k"],
                n=self.rkc["n"],
                p=self.rkc["p"],
                N=self.rkc["n"] - 32,
                P=self.rkc["p"] - 32,
                t=self.rkc["t"],
                z=self.rkc["z"],
                T=self.rkc["t"] - 32,
                Z=self.rkc["z"] - 32,
                s=self.rkc["s"],
                h=self.rkc["h"],
                H=self.rkc["h"] - 32,
            )
        else:
            text = _helptext3.format(
                file_a=self.file_a,
                color_a=self.color_a,
                file_b=self.file_b,
                color_b=self.color_b,
                file_c=self.file_c,
                color_c=self.color_c,
                color_d=self.color_d,
                color_e=self.color_e,
                color_f=self.color_f,
                w=self.rkc["w"],
                x=self.rkc["x"],
                q=self.rkc["q"],
                a=self.rkc["a"],
                b=self.rkc["b"],
                c=self.rkc["c"],
                d=self.rkc["d"],
                e=self.rkc["e"],
                f=self.rkc["f"],
                g=self.rkc["g"],
                A=self.rkc["a"] - 32,
                B=self.rkc["b"] - 32,
                C=self.rkc["c"] - 32,
                D=self.rkc["d"] - 32,
                E=self.rkc["e"] - 32,
                F=self.rkc["f"] - 32,
                G=self.rkc["g"] - 32,
                m=self.rkc["m"],
                edit_cmd=self.edit_cmd,
                M=self.rkc["m"] - 32,
                j=self.rkc["j"],
                k=self.rkc["k"],
                n=self.rkc["n"],
                p=self.rkc["p"],
                N=self.rkc["n"] - 32,
                P=self.rkc["p"] - 32,
                t=self.rkc["t"],
                z=self.rkc["z"],
                T=self.rkc["t"] - 32,
                Z=self.rkc["z"] - 32,
                s=self.rkc["s"],
                h=self.rkc["h"],
                H=self.rkc["h"] - 32,
            )
        return text



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai:
"""
Script to test diff2 and diff3 logic

The diff3 logic is build on top of diff2 logic.

Current diff2 uses Python standard library difflib which uses a variant of
longest contiguous matching subsequence algorithm by Ratcliff and Obershelp
developed in the late 1980's.  If I update this imediff program to use more
modern algorithm, this test may yield slightly different result in some
corner cases.
"""

import difflib
import imediff.diff3lib


def diff23(a, b, c, name=""):
    print("# Test type: " + name)
    print("  a='%s' -> %i" % (a, len(a)))
    print("  b='%s' -> %i" % (b, len(b)))
    print("  c='%s' -> %i" % (c, len(c)))
    sa = difflib.SequenceMatcher(None, a, b)
    print("$ diff2 A B")
    for tag, i1, i2, j1, j2 in sa.get_opcodes():
        print(
            (
                "  %s     a[%d:%d] (%s) / b[%d:%d] (%s)"
                % (tag, i1, i2, a[i1:i2], j1, j2, b[j1:j2])
            )
        )
    print("$ diff2 C B")
    sc = difflib.SequenceMatcher(None, c, b)
    for tag, i1, i2, j1, j2 in sc.get_opcodes():
        print(
            (
                "  %s     c[%d:%d] (%s) / b[%d:%d] (%s)"
                % (tag, i1, i2, c[i1:i2], j1, j2, b[j1:j2])
            )
        )
    s = imediff.diff3lib.SequenceMatcher3(a, b, c, 0, None, True)
    print("$ diff3 A B C")
    for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
        print(
            (
                "  %s     a[%d:%d] (%s) / b[%d:%d] (%s) / c[%d:%d] (%s)"
                % (tag, i1, i2, a[i1:i2], j1, j2, b[j1:j2], k1, k2, c[k1:k2])
            )
        )
    print("===========================================================")
    return


diff23("12345", "12345", "12345", "same")

diff23("a12345z", "1245", "1245", "add a side")

diff23("a12345z", "1245", "a12345z", "add same")

diff23("1245", "1245", "a12345z", "add c side")

diff23("24", "12345", "12345", "delete a side")

diff23("24", "12345", "24", "delete same")

diff23("12345", "12345", "24", "delete c side")

diff23("a2b4c", "12345", "x2y4z", "conflict both ends")

diff23("1a2b4c", "12345", "x2y45z", "conflict skewed")



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - an Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen tool

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018       Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""
from imediff.__main__ import *



#!/usr/bin/python3
# vim:se tw=0 sts=4 ts=4 et ai:

"""
Script to test imediff initializing code
"""
import imediff
import configparser

# print("VERSION={}".format(imediff.VERSION))
# print("PACKAGE={}".format(imediff.PACKAGE))
# print(">>> imediff >>>>>>> key -> value >>>>>>>>>>>>>>>>")
# for k, v in imediff.__dict__.items():
#    print(">>> '{}' -> '{}'".format(k, v))
##########################################################################
args = imediff.initialize_args()
print(">>> args >>>>>>> key -> value >>>>>>>>>>>>>>>>")
# for k, v in args.__dict__.items():
#    print(">>> args >>> '{}' -> '{}'".format(k, v))

##########################################################################
confs = imediff.initialize_confs(args.conf)
print(">>> confs['section'] >>>>>>> 'key' = 'value' >>>>>>>>>>>>>>>>")
for section in confs.sections():
    for key, value in confs[section].items():
        print(">>> confs['{}'] >>> '{}' = '{}'".format(section, key, value))

##########################################################################
print(">>> confs_internal['section'] >>>>>>> 'key' = 'value' >>>>>>>>>>>>>>>>")
confs_internal = configparser.ConfigParser(inline_comment_prefixes=(";", "#"))
confs_internal.read_string(imediff.config_template)
for section in confs_internal.sections():
    for key, value in confs_internal[section].items():
        print(">>> confs['{}'] >>> '{}' = '{}'".format(section, key, value))

##########################################################################
list_a = imediff.read_lines(args.file_a)
list_b = imediff.read_lines(args.file_b)
list_c = imediff.read_lines(args.file_c)
print(">>> list_a >>>>>>> key -> value >>>>>>>>>>>>>>>>")
for i, f in enumerate(list_a):
    print(">>> list_a[{}]='{}'".format(i, f[:-1]))
print(">>> list_b >>>>>>> key -> value >>>>>>>>>>>>>>>>")
for i, f in enumerate(list_b):
    print(">>> list_b[{}]='{}'".format(i, f[:-1]))
print(">>> list_c >>>>>>> key -> value >>>>>>>>>>>>>>>>")
for i, f in enumerate(list_c):
    print(">>> list_c[{}]='{}'".format(i, f[:-1]))
print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")



#!/usr/bin/python3
# vim:se tw=78 sw=4 sts=4 ts=4 et ai si ft=python fileencoding=utf-8 :

"""
Module lines2lib -- matching 2 lines library

Copyright (C) 2021       Osamu Aoki <osamu@debian.org>

"""

import re
from difflib import SequenceMatcher

# utility functions
from imediff.utils import logger, error_exit


class LineMatcher:
    """
    Linematcher

    A public class to help manage 2 lists of similar lines by finding
    matching lines including partial matches while ignoring leading and
    tailing whitespaces.

    Example:
    >>> a = [   "line 1 abcde\\n",
    ...         "line 2 qazws\\n",
    ...         "line 3 edcrf\\n",
    ...         "line 4 tgbyh\\n",
    ...         "line 5 yhnuj\\n",
    ...         "line 6 ujmik\\n",
    ...         "line 7 olpvb\\n",
    ...         "line 8 abcde\\n",
    ...         "line 9 olpvb\\n",
    ...         "line 0 abcde\\n",
    ...         "line 0 abcde\\n",
    ...         "line A olpvb\\n",
    ...         "line B abcde\\n",
    ...         "line Z abcde\\n",
    ...     ]
    >>> b = [   "line 1 abcde\\n",
    ...         " line 2 qazws XXX\\n",
    ...         "  line 3 e dcrf\\n",
    ...         "  line 4 tgbyh\\n",
    ...         "   line 5  yhnuj\\n",
    ...         "   line 6  uj mik\\n",
    ...         "line 7 olpvb\\n",
    ...         "   line 8 abc\\n",
    ...         "   line X XXX\\n",
    ...         "   line X XXX\\n",
    ...         "ine  9 olp vb\\n",
    ...         " l ine  0 abc\\n",
    ...         " line A  olpv b\\n",
    ...         "B  abc\\n",
    ...         "l i n e Z 'a b c d' \\\"e\\n",
    ...     ]
    >>> lines = LineMatcher(a, b)
    >>> lines.print_opcodes()
    match: 0 -> 0, tag = E
        a: line 1 abcde
        b: line 1 abcde
    match: 1 -> 1, tag = N
        a: line 2 qazws
        b:  line 2 qazws XXX
    match: 2 -> 2, tag = F
        a: line 3 edcrf
        b:   line 3 e dcrf
    match: 3 -> 3, tag = F
        a: line 4 tgbyh
        b:   line 4 tgbyh
    match: 4 -> 4, tag = F
        a: line 5 yhnuj
        b:    line 5  yhnuj
    match: 5 -> 5, tag = F
        a: line 6 ujmik
        b:    line 6  uj mik
    match: 6 -> 6, tag = E
        a: line 7 olpvb
        b: line 7 olpvb
    match: 7 -> 7, tag = F
        a: line 8 abcde
        b:    line 8 abc
    UNmatch: 8:8 -> 8:10, tag = N
    match: 8 -> 10, tag = F
        a: line 9 olpvb
        b: ine  9 olp vb
    match: 9 -> 11, tag = F
        a: line 0 abcde
        b:  l ine  0 abc
    UNmatch: 10:11 -> 12:12, tag = N
    match: 11 -> 12, tag = F
        a: line A olpvb
        b:  line A  olpv b
    match: 12 -> 13, tag = N
        a: line B abcde
        b: B  abc
    match: 13 -> 14, tag = F
        a: line Z abcde
        b: l i n e Z 'a b c d' "e
    """

    def __init__(
        self,
        a=[],
        b=[],
        linerule=2,
    ):
        """
        Construct a _LineMatcher

        """

        # initialize
        self.a = a
        self.b = b
        if not (linerule >= 0 and linerule < 20):
            error_exit("E: linerule should be between 0 and 19 but {}".format(linerule))
        # linerule:
        # 0      r""        -- drop none between text, but strip
        # 1      r"\s+"     -- drop all whitespaces
        # 2      r"[\s\"']" -- drop all whitespaces and quotes (default)
        # 3      r"\W+"     -- drop all non-alphanumerics
        # 10     r""        -- drop none between text, but strip and lowercase
        # 11     r"\s+"     -- drop all whitespaces and lowercase
        # 12     r"[\s\"']" -- drop all whitespaces and quotes and lowercase
        # 13     r"\W+"     -- drop all non-alphanumerics and lowercase
        if (linerule % 10) == 0:
            re_preform = re.compile(r"")
        elif (linerule % 10) == 1:
            re_preform = re.compile(r"\s+")
        elif (linerule % 10) == 2:
            re_preform = re.compile(r"[\s\"']+")
        elif (linerule % 10) == 3:
            re_preform = re.compile(r"\W+")
        else:
            re_preform = re.compile(r"")
        self.a_int = []
        for ax in a:
            if linerule < 10:
                filtered_ax = re_preform.sub("", ax).strip()
            else:
                filtered_ax = re_preform.sub("", ax).strip().lower()
            self.a_int.append(filtered_ax)
        self.b_int = []
        for bx in b:
            if linerule < 10:
                filtered_bx = re_preform.sub("", bx).strip()
            else:
                filtered_bx = re_preform.sub("", bx).strip().lower()
            self.b_int.append(filtered_bx)
        self.int = _LineMatcher(
            self.a_int, self.b_int, 0, len(self.a_int), 0, len(self.b_int)
        )

    def get_opcodes(self):
        match = []
        for tag, i1, i2, j1, j2 in self.int.get_opcodes():
            if tag == "E":
                if self.a[i1] == self.b[j1]:
                    tag = "E"
                else:
                    tag = "F"
            match.append((tag, i1, i2, j1, j2))
        return match

    def print_opcodes(self):
        """ """
        for tag, i1, i2, j1, j2 in self.get_opcodes():
            if (i1 + 1) == i2 and (j1 + 1) == j2:
                print("match: {} -> {}, tag = {}".format(i1, j1, tag))
                print("    a: {}".format(self.a[i1]).rstrip())
                print("    b: {}".format(self.b[j1]).rstrip())
            else:
                print("UNmatch: {}:{} -> {}:{}, tag = {}".format(i1, i2, j1, j2, tag))


class _LineMatcher:
    """
    _LineMatcher

    A private class to help manage 2 lists of similar lines by finding
    matching lines including partial matches.

    Example:
    >>> a = [   "line1abcde",
    ...         "line2qazws",
    ...         "line3edcrf",
    ...         "line4tgbyh",
    ...         "line5yhnuj",
    ...         "line6ujmik",
    ...         "line7olpvb",
    ...         "line8abcde",
    ...         "line9olpvb",
    ...         "line0abcde",
    ...         "line0abcde",
    ...         "lineAolpvb",
    ...         "lineBabcde",
    ...         "lineZabcde",
    ...     ]
    >>> b = [   "line1abcde",
    ...         "line2qazwsx",
    ...         "line3edcrf",
    ...         "line4tgbyhx",
    ...         "line5yhnuj",
    ...         "line6ujmikz",
    ...         "line7olpvb",
    ...         "line8abcdez",
    ...         "line9olpvb",
    ...         "line0abcde",
    ...         "line0abcdexx",
    ...         "lineAolpvb",
    ...         "lineBabcde",
    ...         "lineZabcdex",
    ...     ]
    >>> lines = _LineMatcher(a, b, 0, len(a), 0, len(b))
    >>> lines.print_opcodes()
    match: 0 -> 0, tag = E
        a: line1abcde
        b: line1abcde
    match: 1 -> 1, tag = N
        a: line2qazws
        b: line2qazwsx
    match: 2 -> 2, tag = E
        a: line3edcrf
        b: line3edcrf
    match: 3 -> 3, tag = N
        a: line4tgbyh
        b: line4tgbyhx
    match: 4 -> 4, tag = E
        a: line5yhnuj
        b: line5yhnuj
    match: 5 -> 5, tag = N
        a: line6ujmik
        b: line6ujmikz
    match: 6 -> 6, tag = E
        a: line7olpvb
        b: line7olpvb
    match: 7 -> 7, tag = N
        a: line8abcde
        b: line8abcdez
    match: 8 -> 8, tag = E
        a: line9olpvb
        b: line9olpvb
    match: 9 -> 9, tag = E
        a: line0abcde
        b: line0abcde
    match: 10 -> 10, tag = N
        a: line0abcde
        b: line0abcdexx
    match: 11 -> 11, tag = E
        a: lineAolpvb
        b: lineAolpvb
    match: 12 -> 12, tag = E
        a: lineBabcde
        b: lineBabcde
    match: 13 -> 13, tag = N
        a: lineZabcde
        b: lineZabcdex
    """

    def __init__(
        self,
        a=[],
        b=[],
        is1=0,
        is2=0,
        js1=0,
        js2=0,
        depth=0,
        length=128,
        lenmin=1,
        factor=8,
    ):
        """
        Construct a _LineMatcher

        """

        # initialize
        self.a = a
        self.b = b
        self.is1 = is1
        self.is2 = is2
        self.js1 = js1
        self.js2 = js2
        self.depth = depth
        self.lenmin = lenmin
        self.factor = factor
        maxlen = 0
        for i in range(is1, is2):
            len_a = len(a[i])
            if len_a > maxlen:
                maxlen = len_a
        for j in range(js1, js2):
            len_b = len(b[j])
            if len_b > maxlen:
                maxlen = len_b
        if length >= (maxlen // 2):
            length = maxlen // 2
        self.length = length

    def get_opcodes(self):
        if self.depth == 0:  # depth = 0
            side = 0
            logger.debug(
                "===  a[{}:{}]/b[{}:{}]  ===  d={:02d}  ===  line[:]  ===".format(
                    self.is1, self.is2, self.js1, self.js2, self.depth
                )
            )
        elif self.depth > 0:  # depth > 0
            if self.depth % 2 == 1:  # depth = 1, 3, 5, ...
                side = +1
                logger.debug(
                    "===  a[{}:{}]/b[{}:{}]  ===  d={:02d}  ===  line[:{:02d}] ===".format(
                        self.is1, self.is2, self.js1, self.js2, self.depth, self.length
                    )
                )
            else:  # self.depth % 2 == 0:  # depth = 2, 4, 6, ...
                side = -1
                logger.debug(
                    "===  a[{}:{}]/b[{}:{}]  ===  d={:02d}  ===  line[-{:02d}:] ===".format(
                        self.is1, self.is2, self.js1, self.js2, self.depth, self.length
                    )
                )
        else:
            error_exit(
                "===  a[{}:{}]/b[{}:{}]  ===  d={:02d} should be non-negative".format(
                    self.is1, self.is2, self.js1, self.js2, self.depth
                )
            )
        if side == 0:
            # self.is1, self.is2, self.js1, self.js2 are known to cover all
            am = self.a
            bm = self.b
        elif side == 1:  # left side match
            am = []
            bm = []
            for i in range(self.is1, self.is2):
                am.append(self.a[i][: self.length])
            for j in range(self.js1, self.js2):
                bm.append(self.b[j][: self.length])
        else:  # side == -1, right side match
            am = []
            bm = []
            for i in range(self.is1, self.is2):
                am.append(self.a[i][-self.length :])
            for j in range(self.js1, self.js2):
                bm.append(self.b[j][-self.length :])
        for i in range(self.is1, self.is2):
            logger.debug(
                "filter a[{}] -> am[{}]='{}'".format(i, i - self.is1, am[i - self.is1])
            )
        for j in range(self.js1, self.js2):
            logger.debug(
                "filter b[{}] -> bm[{}]='{}'".format(j, j - self.js1, bm[j - self.js1])
            )
        seq = SequenceMatcher(None, am, bm)
        match = []
        for tag, i1, i2, j1, j2 in seq.get_opcodes():
            logger.debug(
                ">>> tag={}  ===  a[{}:{}]/b[{}:{}]".format(
                    tag, i1 + self.is1, i2 + self.is1, j1 + self.js1, j2 + self.js1
                )
            )
            if tag == "equal":
                for i in range(i1, i2):
                    ip = self.is1 + i
                    jp = self.js1 + j1 + i - i1
                    if side == 0:
                        # full match on filtered line
                        match.append(("E", ip, ip + 1, jp, jp + 1))
                    else:
                        # partial match only (F for fuzzy)
                        match.append(("F", ip, ip + 1, jp, jp + 1))
            elif i1 == i2 or j1 == j2:
                # clean insert/delete
                match.append(
                    ("N", i1 + self.is1, i2 + self.is1, j1 + self.js1, j2 + self.js1)
                )
            elif (i1 + 1) == i2 and (j1 + 1) == j2:
                # single line change -> assume fuzzy match without checking
                match.append(
                    ("F", i1 + self.is1, i2 + self.is1, j1 + self.js1, j2 + self.js1)
                )
            else:
                # dig deeper for multi-line changes
                if side == 0:
                    # full -> left side
                    match.extend(
                        _LineMatcher(
                            a=self.a,
                            b=self.b,
                            is1=i1 + self.is1,
                            is2=i2 + self.is1,
                            js1=j1 + self.js1,
                            js2=j2 + self.js1,
                            length=self.length,
                            depth=self.depth + 1,
                        ).get_opcodes()
                    )
                elif side == +1:
                    # left side -> right side
                    match.extend(
                        _LineMatcher(
                            a=self.a,
                            b=self.b,
                            is1=i1 + self.is1,
                            is2=i2 + self.is1,
                            js1=j1 + self.js1,
                            js2=j2 + self.js1,
                            length=self.length,
                            depth=self.depth + 1,
                        ).get_opcodes()
                    )
                elif self.length > self.lenmin:  # side == -1
                    # right side -> left side (shorter)
                    match.extend(
                        _LineMatcher(
                            a=self.a,
                            b=self.b,
                            is1=i1 + self.is1,
                            is2=i2 + self.is1,
                            js1=j1 + self.js1,
                            js2=j2 + self.js1,
                            length=self.length * self.factor // 10,
                            depth=self.depth + 1,
                        ).get_opcodes()
                    )
                else:
                    # no more shorter, give up as multi-line block change
                    match.append(
                        (
                            "N",
                            i1 + self.is1,
                            i2 + self.is1,
                            j1 + self.js1,
                            j2 + self.js1,
                        )
                    )
            logger.debug(
                "<<< tag={}  ===  a[{}:{}]/b[{}:{}]".format(
                    tag, i1 + self.is1, i2 + self.is1, j1 + self.js1, j2 + self.js1
                )
            )
        return match

    def print_opcodes(self):
        """ """
        for tag, i1, i2, j1, j2 in self.get_opcodes():
            if (i1 + 1) == i2 and (j1 + 1) == j2:
                print("match: {} -> {}, tag = {}".format(i1, j1, tag))
                print("    a: {}".format(self.a[i1]))
                print("    b: {}".format(self.b[j1]))
            else:
                print("UNmatch: {}:{} -> {}:{}, tag = {}".format(i1, i2, j1, j2, tag))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai:
"""
Test many aspect of imediff package via unittest.

To test the in-source-tree module, invoke this script in this directory as:

 $ export PYTHONPATH=../src
 $ python3 test_diff23lib.py -v

To test the system installed module, invoke this script directly in this
directory as:

 $ python3 test_diff23lib.py -v

Current diff2 uses Python standard library difflib which uses a variant of
longest contiguous matching subsequence algorithm by Ratcliff and Obershelp
developed in the late 1980's.  If I update this imediff program to use more
modern algorithm, this test may yield slightly different result.
"""
import unittest
import subprocess
import os
import os.path
import imediff.diff3lib

# path to test directory
pwd = os.path.dirname(os.path.abspath(__file__))


class TestImediff(unittest.TestCase):

    a = "a12b345c6789d"
    b = "123456789"
    c = "a1234b567c89d"

    def test_diff3lib_abc(self):
        a = "a12b345c6789d"
        b = "123456789"
        c = "a1234b567c89d"
        self.assertEqual(
            imediff.diff3lib.SequenceMatcher3(a, b, c, 0, None, True).get_opcodes(),
            [
                ("e", 0, 1, 0, 0, 0, 1),
                ("E", 1, 3, 0, 2, 1, 3),
                ("A", 3, 4, 2, 2, 3, 3),
                ("E", 4, 6, 2, 4, 3, 5),
                ("C", 6, 6, 4, 4, 5, 6),
                ("E", 6, 7, 4, 5, 6, 7),
                ("A", 7, 8, 5, 5, 7, 7),
                ("E", 8, 10, 5, 7, 7, 9),
                ("C", 10, 10, 7, 7, 9, 10),
                ("E", 10, 12, 7, 9, 10, 12),
                ("e", 12, 13, 9, 9, 12, 13),
            ],
        )
        return

    def test_diff3lib_doctest(self):
        #            "cd " + pwd + "; pwd; ls -lR",
        result = subprocess.call(
            "cd " + pwd + "; ./diff3lib.py",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_diff23(self):
        result = subprocess.call(
            "cd " + pwd + ";./_diff23.py >z_diff23.out", shell=True
        )
        result = subprocess.call(
            "cd " + pwd + ";diff z_diff23.out z_diff23.ref >/dev/null", shell=True
        )
        self.assertEqual(result, 0)
        return

    def test_imediff2(self):
        result = subprocess.call(
            "cd " + pwd + ";./_imediff.py  -C BOGUS -n file_a file_b -o z_imediff2.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + pwd + ";diff z_imediff2.out z_imediff2.ref >/dev/null", shell=True
        )
        self.assertEqual(result, 0)
        return

    def test_imediff3(self):
        result = subprocess.call(
            "cd "
            + pwd
            + ";./_imediff.py -C BOGUS -n file_a file_b file_c -o z_imediff3.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + pwd + ";diff z_imediff3.out z_imediff3.ref >/dev/null", shell=True
        )
        self.assertEqual(result, 0)
        return

    def test_initialize(self):
        result = subprocess.call(
            "cd " + pwd + ";./_initialize.py  > z_initialize.out", shell=True
        )
        result = subprocess.call(
            "cd " + pwd + ";diff z_initialize.out z_initialize.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def test_macrotrans(self):
        result = subprocess.call(
            "cd "
            + pwd
            + ";./_macrotrans.py -C alternative_imediff.conf -n file_a file_b file_c > z_macrotrans.out",
            shell=True,
        )
        result = subprocess.call(
            "cd " + pwd + ";diff z_macrotrans.out z_macrotrans.ref >/dev/null",
            shell=True,
        )
        self.assertEqual(result, 0)
        return

    def tearDown(self):
        result = subprocess.call("cd " + pwd + ";rm -f z_*.out", shell=True)
        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - an Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen tool

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018-2021  Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""
import tempfile
import os
import io
import time

from difflib import SequenceMatcher
from imediff.utils import error_exit, logger, read_lines, write_file
from imediff.lines2lib import LineMatcher
from imediff.diff3lib import SequenceMatcher3


class TextData:  # Non-TUI data
    """Non curses class to handle diff data for 2 or 3 lines"""

    #     i: index for self.opcodes
    #     j: index for self.actives (non-persistent)
    #     self.active: index for self.actives (persistent)
    def __init__(self, list_a, list_b, list_c, args, confs):
        self.diff_mode = args.diff_mode
        self.default_mode = args.default_mode
        self.file_a = args.file_a
        self.file_b = args.file_b
        self.file_c = args.file_c
        self.file_o = args.output
        self.list_a = list_a
        self.list_b = list_b
        self.list_c = list_c
        self.sloppy = args.sloppy
        self.isjunk = args.isjunk
        self.linerule = args.linerule
        self.edit_cmd = args.edit_cmd
        self.macro = args.macro
        new_mode = self.default_mode
        row = None  # used only by TUI to match index of opcodes to textpad row
        bf = None  # used only by TUI to store the editor result
        # set from confs
        if confs["config"]["confirm_exit"] != "False":
            self.confirm_exit = True
        else:
            self.confirm_exit = False
        if confs["config"]["confirm_quit"] != "False":
            self.confirm_quit = True
        else:
            self.confirm_quit = False
        self.ls0 = confs["line_separator"]["ls0"] + " %s\n"
        self.ls1 = confs["line_separator"]["ls1"] + " %s\n"
        self.ls2 = confs["line_separator"]["ls2"] + "\n"
        self.ls3 = confs["line_separator"]["ls3"] + " %s\n"
        self.ws0 = confs["word_separator"]["ws0"]
        self.ws1 = confs["word_separator"]["ws1"]
        self.ws2 = confs["word_separator"]["ws2"]
        self.ws3 = confs["word_separator"]["ws3"]
        # command key translation
        self.kc = dict()  # customized key code to original key code
        self.rkc = dict()  # original key chr to customized key chr
        for cmd_name, key_char in confs["key"].items():
            self.kc[ord(key_char[:1])] = ord(cmd_name[-1:])
            self.rkc[cmd_name[-1:]] = ord(key_char[:1])
            logger.debug(
                "[key] section: left_side='{}' right_side='{}' : self.krc[{}] = '{}'".format(
                    cmd_name, key_char, cmd_name[-1:], self.rkc[cmd_name[-1:]]
                )
            )
        # parse input data
        if self.diff_mode == 2:
            sequence = LineMatcher(list_a, list_b)
            opcodes = sequence.get_opcodes()
            k1 = k2 = None
            # Set initial mode to "a" or "d"
            self.opcodes = [
                (tag, i1, i2, j1, j2, k1, k2, "a" if tag == "E" else "d", row, bf)
                for (tag, i1, i2, j1, j2) in opcodes
            ]
            self.actives = [
                j for j, (tag, i1, i2, j1, j2) in enumerate(opcodes) if tag != "E"
            ]
        else:
            sequence = SequenceMatcher3(list_a, list_b, list_c, 1)
            opcodes = sequence.get_opcodes()
            # Set initial mode to "a" or "d"
            self.opcodes = [
                (tag, i1, i2, j1, j2, k1, k2, "a" if tag in "Ee" else "d", row, bf)
                for (tag, i1, i2, j1, j2, k1, k2) in opcodes
            ]
            self.actives = [
                j
                for j, (tag, i1, i2, j1, j2, k1, k2) in enumerate(opcodes)
                if tag not in "Ee"
            ]
        # self.actives[i] = j -> self.rev_actives[j] = i
        self.rev_actives = dict()
        for i, j in enumerate(self.actives):
            if new_mode != "d":
                self.set_mode(j, new_mode)
            self.rev_actives[j] = i
        if len(self.actives) == 0:
            self.active = None
        else:
            self.active = 0
        self.active_old = self.active
        self.update_active = True
        self.update_textpad = True
        # save memory
        del opcodes  # this is not "self.opcodes"
        del sequence
        return

    def merge_diff(self, i):
        """Return content for diff by line"""
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        logger.debug(
            "merge_wdiff: tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] mode={}, row={}, bf='{}'".format(
                tag, i1, i2, j1, j2, k1, k2, mode, row, bf
            )
        )
        content = list()
        content += [self.ls0 % self.file_a]
        content += self.list_a[i1:i2]
        if self.diff_mode == 2:
            content += [self.ls2]
            content += self.list_b[j1:j2]
            content += [self.ls3 % self.file_b]
        else:  # self.diff_mode == 3
            content += [self.ls1 % self.file_b]
            content += self.list_b[j1:j2]
            content += [self.ls2]
            content += self.list_c[k1:k2]
            content += [self.ls3 % self.file_c]
        return content

    def merge_wdiff2(self, i):
        """Return content for wdiff by line (2 files)"""
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        logger.debug(
            "merge_wdiff2: tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] mode={}, row={}, bf='{}'".format(
                tag, i1, i2, j1, j2, k1, k2, mode, row, bf
            )
        )
        line_a = "".join(self.list_a[i1:i2])
        line_b = "".join(self.list_b[j1:j2])
        if self.isjunk:
            isjunk = None
        else:
            isjunk = lambda x: x in " \t"
        seq = SequenceMatcher(isjunk, line_a, line_b, None)
        opcodes = seq.get_opcodes()
        line = ""
        for tag, wi1, wi2, wj1, wj2 in opcodes:
            if tag == "equal":
                line += line_a[wi1:wi2]
            else:  # other tags
                line += self.ws0
                line += line_a[wi1:wi2]
                line += self.ws1
                line += line_b[wj1:wj2]
                line += self.ws3
        content = line.splitlines(keepends=True)
        return content

    def merge_wdiff3(self, i):
        """Return content for wdiff by line (3 files)"""
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        logger.debug(
            "merge_wdiff3: tag={} a[{}:{}]/b[{}:{}]/c[{}:{}] mode={}, row={}, bf='{}'".format(
                tag, i1, i2, j1, j2, k1, k2, mode, row, bf
            )
        )
        line_a = "".join(self.list_a[i1:i2])
        line_b = "".join(self.list_b[j1:j2])
        line_c = "".join(self.list_c[k1:k2])
        if self.isjunk:
            isjunk = None
        else:
            isjunk = lambda x: x in " \t"
        wseq = SequenceMatcher3(line_a, line_b, line_c, 0, isjunk, True)
        wopcodes = wseq.get_opcodes()
        # logger.debug("wdiff3: \nwopcodesc >>>>> {}".format(wopcodes))
        line = ""
        clean_merge = True
        for tag, wi1, wi2, wj1, wj2, wk1, wk2 in wopcodes:
            if tag == "E" or tag == "e" or tag == "A":
                line += line_a[wi1:wi2]
            elif tag == "C":
                line += line_c[wk1:wk2]
            else:  # tag == "N"
                clean_merge = False
                line += self.ws0
                line += line_a[wi1:wi2]
                line += self.ws1
                line += line_b[wj1:wj2]
                line += self.ws2
                line += line_c[wk1:wk2]
                line += self.ws3
        content = line.splitlines(keepends=True)
        return (clean_merge, content)

    def get_tag(self, i):
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        return tag

    def get_mode(self, i):
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        return mode

    def get_row(self, i):
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        return row

    def get_bf(self, i):
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        return bf

    def get_content(self, i):
        """Return content based on mode"""
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        if tag == "E" or tag == "e":
            content = self.list_a[i1:i2]
        elif mode == "a":
            content = self.list_a[i1:i2]
        elif mode == "b":
            content = self.list_b[j1:j2]
        elif mode == "c":
            content = self.list_c[k1:k2]
        elif mode == "d":
            content = self.merge_diff(i)
        elif mode == "e":
            if bf is not None:
                content = bf
            else:
                error_exit("Bad mode='e' with missing edited buffer text\n")
        elif mode == "f":
            if self.diff_mode == 2:
                content = self.merge_wdiff2(i)
            else:  # self.diff_mode == 3
                (clean_merge, content) = self.merge_wdiff3(i)
        elif mode == "g":
            if self.diff_mode == 2:
                content = self.merge_diff(i)
            else:  # self.diff_mode == 3
                (clean_merge, content) = self.merge_wdiff3(i)
        else:
            error_exit("Bad mode='{}'\n".format(mode))
        # content is at least [] (at least empty list)
        if content is None:
            error_exit("content can't be None")
        return content

    def set_mode(self, i, new_mode):
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        if new_mode in "abd":
            mode = new_mode
        elif self.diff_mode == 2:  # mode is always in "abdef"
            if new_mode == "c":
                mode = "d"  # hidden alias
            elif new_mode == "f":
                mode = "f"
            else:  # for new_mode in "eg"
                if bf is not None:
                    mode = "e"
                else:  #  for mode in "abdef"
                    pass
        else:  # self.diff_mode == 3
            if new_mode == "c":
                mode = "c"
            elif new_mode == "f":
                (clean_merge, content) = self.merge_wdiff3(i)
                if clean_merge:
                    mode = "g"
                else:
                    mode = "f"
            elif new_mode == "e":
                if bf is not None:
                    mode = "e"
                else:  # for mode in "abcdefg"
                    pass
            else:  # new_mode == "g":
                if bf is not None:
                    mode = "e"
                elif tag == "A":
                    mode = "a"
                elif tag == "C":
                    mode = "c"
                elif mode in "abceg":
                    pass
                else:  # for mode in "df" and tag == "N"
                    (clean_merge, content) = self.merge_wdiff3(i)
                    if clean_merge:
                        mode = "g"
                    else:
                        pass  # mode to "d" or "f"
        self.opcodes[i] = (tag, i1, i2, j1, j2, k1, k2, mode, row, bf)
        self.update_textpad = True
        return

    def set_all_mode(self, new_mode):
        for i in range(len(self.opcodes)):
            logger.debug("set_all_mode: i={} {}".format(i, new_mode))
            self.set_mode(i, new_mode)
        return

    def set_row(self, i, new_row):  # used by TUI
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        # row is passed by value but chunk is passed by reference !
        self.opcodes[i] = (tag, i1, i2, j1, j2, k1, k2, mode, new_row, bf)
        return

    def set_bf(self, i, new_bf):
        (tag, i1, i2, j1, j2, k1, k2, mode, row, bf) = self.opcodes[i]
        self.opcodes[i] = (tag, i1, i2, j1, j2, k1, k2, mode, row, new_bf)
        return

    def editor(self, i):
        content = self.get_content(i)
        linebuf = self.ext_editor(content)
        self.set_bf(i, linebuf)
        self.set_mode(i, "e")
        return

    def del_editor(self, i):
        self.set_bf(i, None)
        self.set_mode(i, "d")
        return

    def ext_editor(self, content):
        with tempfile.NamedTemporaryFile(
            mode="w",
            buffering=io.DEFAULT_BUFFER_SIZE,
            suffix=".tmp",
            prefix="imediff.",
            dir=".",
            delete=False,
        ) as fp:
            tmpfname = fp.name
            if len(content):
                for line in content:
                    fp.write(line)
        time.sleep(0.1)  # make the change visible
        editor_ret = os.system("%s %s" % (self.edit_cmd, tmpfname))
        time.sleep(0.1)  # make the change visible
        # time.sleep(5.0) # debug
        if editor_ret == 0:
            linebuf = read_lines(tmpfname)
        else:
            linebuf = [""]  # Ignore editor errors
        os.unlink(tmpfname)
        return linebuf

    def get_output(self):
        """Return output of all content"""
        output = ""
        for i in range(len(self.opcodes)):
            content = self.get_content(i)
            output += "".join(content)
        return output

    def active_next(self):
        """Jump to the next active chunk"""
        self.active_old = self.active
        if self.active is not None:
            self.active = min(self.active + 1, len(self.actives) - 1)
        if self.active_old != self.active:
            self.update_active = True
        else:
            self.row += 1
        return

    def active_prev(self):
        """Jump to the previous active chunk"""
        self.active_old = self.active
        if self.active is not None:
            self.active = max(self.active - 1, 0)
        if self.active_old != self.active:
            self.update_active = True
        return

    def active_home(self):
        """Jump to the first active chunk"""
        self.active_old = self.active
        if self.active is not None:
            self.active = 0
        if self.active_old != self.active:
            self.update_active = True
        return

    def active_end(self):
        """Jump to the last active chunk"""
        self.active_old = self.active
        if self.active is not None:
            self.active = len(self.actives) - 1
        if self.active_old != self.active:
            self.update_active = True
        else:
            self.row += 1
        return

    def get_unresolved_count(self):
        """Count 'd' or 'f' mode in active chunk"""
        count = 0
        for j, i in enumerate(self.actives):
            if self.get_mode(i) in "df":
                count += 1
        return count

    def diff_next(self):
        """Jump to the next diff chunk"""
        self.active_old = self.active
        active = None
        for j in range(self.active + 1, len(self.actives)):
            i = self.actives[j]
            if self.get_mode(i) in "df":
                active = j
                break
        if active is not None:
            self.active = active
        if self.active_old != self.active:
            self.update_active = True
        else:
            self.row += 1
        return

    def diff_prev(self):
        """Jump to the previous diff chunk"""
        self.active_old = self.active
        active = None
        for j in range(self.active - 1, -1, -1):
            i = self.actives[j]
            if self.get_mode(i) in "df":
                active = j
                break
        if active is not None:
            self.active = active
        if self.active_old != self.active:
            self.update_active = True
        return

    def diff_home(self):
        """Jump to the first diff chunk"""
        self.active_old = self.active
        active = None
        for j in range(0, self.active):
            i = self.actives[j]
            if self.get_mode(i) in "df":
                active = j
                break
        if active is not None:
            self.active = active
        if self.active_old != self.active:
            self.update_active = True
        return

    def diff_end(self):
        """Jump to the last diff chunk"""
        self.active_old = self.active
        active = None
        for j in range(len(self.actives) - 1, self.active, -1):
            i = self.actives[j]
            if self.get_mode(i) in "df":
                active = j
                break
        if active is not None:
            self.active = active
        if self.active_old != self.active:
            self.update_active = True
        else:
            self.row += 1
        return

    def c_translated(self, c):
        """translate keys in (' '...'~') according to ~/.imediff"""
        if c >= ord("A") and c <= ord("Z"):
            # special handling of upper case letters
            if c + 32 in list(self.kc.keys()):
                c = self.kc[c + 32] - 32
            logger.debug("c_translated chr = '{}'".format(chr(c)))
        elif c >= ord(" ") and c <= ord("~"):
            if c in list(self.kc.keys()):
                c = self.kc[c]
            logger.debug("c_translated chr = '{}'".format(chr(c)))
        else:
            logger.debug("c_translated num = '{}'".format(c))
        return c

    def getch_translated(self):  # overridden for TUI by subclassing
        """Macro parsing instead of curses getch"""
        if len(self.macro):
            c = ord(self.macro[:1])
            c = self.c_translated(c)
            self.macro = self.macro[1:]
        else:
            c = 0  # End of MACRO
        return c

    def command_loop(self):  # overridden for TUI by subclassing
        """Non-interactive driven by MACRO"""
        logger.debug("command-loop start - macro")
        while True:
            # reset flags
            self.update_textpad = False  # TUI
            self.update_active = False
            c = self.getch_translated()
            if c > ord(" ") and c < 127:
                ch = chr(c)
            else:
                ch = " "

            if c == 0 or ch == "w" or ch == "x":
                output = self.get_output()
                write_file(self.file_o, output)
                break
            elif ch == "q":
                # No prompt for CLI
                break
            if self.active is not None:
                # get active chunk
                # Explicitly select chunk mode
                if ch in "abcdefg":
                    self.set_mode(self.actives[self.active], ch)
                elif ch in "1234567":
                    self.set_mode(
                        self.actives[self.active], chr(ord(ch) - ord("1") + ord("a"))
                    )
                elif ch in "ABCDEFG":
                    self.set_all_mode(ch.lower())
                elif ch == "m":
                    self.editor(self.actives[self.active])
                elif ch == "M":
                    self.del_editor(self.actives[self.active])
                elif ch == "n":
                    self.active_next()
                elif ch == "p":
                    self.active_prev()
                elif ch == "t":
                    self.active_home()
                elif ch == "z":
                    self.active_end()
                elif ch == "N":
                    self.diff_next()
                elif ch == "P":
                    self.diff_prev()
                elif ch == "T":
                    self.diff_home()
                elif ch == "Z":
                    self.diff_end()
                else:
                    pass
            else:
                pass
        logger.debug("command-loop")
        return



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - an Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen tool

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018       Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

import logging
import errno
import io
import sys
import traceback
import curses
import unicodedata


__all__ = ["_", "console_width", "logger", "read_lines", "error_exit", "write_file"]

# Utility functions

# gettext for i18n (commented out since gettext is not used)
# gettext.bindtextdomain("imediff")
# gettext.textdomain("imediff")
# _ = gettext.gettext

# Readable version of: _ = lambda x : x
def _(x):
    return x

# Console width with Zenkaku=2, Hankaku=1 (Hankaku=ASCII etc.)
# Latin-1, CJK focus simplification applied.
# (For Arabic, Hebrew etc., we need some refinement)
def console_width(text):
    width = 0
    for c in text:
        if c == "\t":  # TAB = 8
            width += 8
        elif unicodedata.east_asian_width(c) in "FWA":  # Zenkaku
            width += 2
        else:
            width += 1
    return width


# logger
logger = logging.getLogger(__name__)

# file read
def read_lines(filename):
    if filename is None or filename == "":
        lines = []
    else:
        try:
            with open(filename, buffering=io.DEFAULT_BUFFER_SIZE) as fp:
                lines = fp.readlines()  # read into list
        except IOError as e:
            (error, message) = e.args
            if error == errno.ENOENT:
                lines = []
            else:
                error_exit("Could not read '{}': {}\n".format(filename, message))
        except:
            error_exit(
                "read_lines: Unexpected error: {} {}".format(
                    sys.exc_info()[0], sys.exc_info()[1]
                )
            )
    return lines


# error exit
def error_exit(msg):
    # time.sleep(5.0)
    try:
        curses.nocbreak()
        curses.echo()
        curses.endwin()
    except curses.error:
        pass
    except:
        error_exit(
            "error_exit: Unexpected error: {} {}".format(
                sys.exc_info()[0], sys.exc_info()[1]
            )
        )
    logger.error(msg)
    trace = traceback.format_exc()
    if trace != "NoneType: None\n":
        logger.error(trace)
    sys.exit(2)


# file output
def write_file(filename, output):
    if filename is None or filename == "-" or filename == "":
        sys.stderr.write(output)
    else:
        try:
            with open(filename, mode="w", buffering=io.DEFAULT_BUFFER_SIZE) as fp:
                fp.write(output)
        except IOError:
            error_exit("Error in creating output file: {}".format(filename))
    return



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
IMEDIFF - an Interactive Merge Editor for DIFF2 and DIFF3
          Curses based single-pane fullscreen tool

Copyright (C) 2003, 2004 Jarno Elonen <elonen@iki.fi>
Copyright (C) 2018       Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

import os
import io
import curses

from imediff.utils import error_exit, logger

# Update version below only when configuration API changes

config_template = """\
# imediff configuration file
# Please edit right side of assignment to customize
# ini-type syntax where # starts comment
# Don't create duplicate definition
# Key side is case insensitive.

[config]
version = 2.3       # DON'T EDIT THIS.  This is for future upgrade tracking.
confirm_exit = True # Set as "False" to save and exit without pause
confirm_quit = True # Set as "False" to quit without pause
#editor = vim       # Set this to override /usr/bin/editor and $EDITOR

[key]               # key assignment for the single key command
select_a = a        # set mode a to select 'a' buffer
select_b = b        # set mode b to select 'b' buffer
select_c = c        # set mode c to select 'c' buffer (diff3)
select_d = d        # set mode d to select diff content
select_e = e        # set mode e to select editor buffer
select_f = f        # set mode f to select wdiff content
select_g = g        # set _g_ood default mode (diff3)
select_h = h        # show _h_elp screen
select_j = j        # move down display scope
select_k = k        # move up display scope
select_m = m        # start editor to _m_odify content
select_n = n        # move active selection to _n_ext
select_p = p        # move active selection to _p_revious
select_q = q        # _q_uit imediff without saving the result
select_s = s        # merge _s_tatus
select_t = t        # move active selection to home=_t_op
select_w = w        # _w_rite result and e_x_it program
select_x = x        # _w_rite result and e_x_it program
select_y = y        # key for "_Y_es" answer
select_z = z        # move active selection to end=_z_

[color_diff2]       # color assignment for imediff with 2 files
color_a = BLUE      # color for mode a  (OLDER)
color_b = RED       # color for mode b  (NEWER)
color_c = MAGENTA   #                   (not used with diff2)
color_d = GREEN     # color for mode d  (DIFF)
color_e = YELLOW    # color for mode e  (EDITOR)
color_f = CYAN      # color for mode f  (WDIFF)

[color_diff3]       # color assignment for imediff with 3 files
color_a = BLUE      # color for mode a  (YOURS)
color_b = MAGENTA   # color for mode b  (OLD COMMON)
color_c = RED       # color for mode c  (THEIRS)
color_d = GREEN     # color for mode d  (DIFF)
color_e = YELLOW    # color for mode e  (EDITOR)
color_f = CYAN      # color for mode f  (WDIFF)

[line_separator]    # diff output formatting strings
                    # diff2 uses       ls0,      ls2, ls3
                    # diff3 uses       ls0, ls1, ls2, ls3
                    # File name  added ls0, ls1,      ls3
ls0 = <<<<<<<
ls1 = |||||||
ls2 = =======
ls3 = >>>>>>>

[word_separator]    # wdiff output formatting strings
                    # wdiff2 uses       ws0, ws1,      ws3
                    # wdiff3 uses       ws0, ws1, ws2, ws3
ws0 = {
ws1 = |
ws2 = |
ws3 = }
# alternative1 for UTF-8 terminal
#ws0 = «🅐
#ws1 = 🅑
#ws2 = 🅒
#ws3 = »
# alternative2 for UTF-8 terminal
#ws0 = «
#ws1 = ╪
#ws2 = ╫
#ws3 = »
"""

cc = dict()
cc["BLUE"] = curses.COLOR_BLUE
cc["RED"] = curses.COLOR_RED
cc["YELLOW"] = curses.COLOR_YELLOW
cc["GREEN"] = curses.COLOR_GREEN
cc["CYAN"] = curses.COLOR_CYAN
cc["MAGENTA"] = curses.COLOR_MAGENTA
cc["WHITE"] = curses.COLOR_WHITE
cc["BLACK"] = curses.COLOR_BLACK


def create_template(conf):
    config_file = os.path.expanduser(conf)
    if not os.path.exists(config_file):
        logger.debug("create configuration file: {}".format(conf))
        try:
            with open(config_file, mode="w", buffering=io.DEFAULT_BUFFER_SIZE) as ofp:
                ofp.write(config_template)
        except IOError:
            error_exit("Error in creating configuration file: {}".format(conf))
    else:
        error_exit("Erase {} before 'imediff -t'".format(conf))
    return


# Generate template file: TEMPLATE.imediff
if __name__ == "__main__":

    create_template("TEMPLATE.imediff")



#!/usr/bin/python3
# vim:se tw=79 sts=4 ts=4 et ai fileencoding=utf-8 :

"""
Module diff3lib -- wrapper for diff3 (generic)

Class SequenceMatcher3:
    A flexible class for comparing 3 of sequences of any type.

Copyright (C) 2018-2021       Osamu Aoki <osamu@debian.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
"""

from difflib import SequenceMatcher
from imediff.lines2lib import LineMatcher
from imediff.utils import logger


class SequenceMatcher3:

    """
    SequenceMatcher3 is a matching class for comparing three of sequences of
    any type, so long as the sequence elements are hashable for matcher=0.

    SequenceMatcher3 is a matching class for comparing three of lists of
    lines for matcher=1.

    Let's call these three sequences as A (yours), B (base), and C (theirs).
    Let's assume A and C are derivatives from B and try to merge C into A.
    The basic rule for the merging of three of sequences is as follows:
        * Check B-A matching and B-C matching situation.
        * Walk through B to see which sides made change. (12 cases)
        * Check if the high end of range is available (Either "E" or deletion)
        * Pick one making change as the merged content.
            * Merged as equal: 3 cases
                BA      EEEENN..    EEEEEE..    EEEENN..
                BC      EEEENN..    EEEENN..    EEEEEE..
                        ^^^^        ^^^^        ^^^^
                Range   AC          AC          AC
                Tag     E           E           E
            * Merged to pick A: 2 cases, undecided to pick conflict 1 case
                BA      NNNNNN..    NNNNEE..                NNNNEE..
                BC      EEEENN..    EEEENN..                EEEEEE..
                        ^^^^        ^^^^                    ^^^^
                Range   A           AC                      AC
                Tag     N*          A                       A
            * Merged to pick C: 2 cases, undecided to pick conflict 1 case
                BA      EEEENN..                EEEENN..    EEEEEE..
                BC      NNNNNN..                NNNNEE..    NNNNEE..
                        ^^^^                    ^^^^        ^^^^
                Range   C                       AC          AC
                Tag     N*                      C           C
            * Merged as conflict: 1 case, undecided to pick conflict 2 cases
                BA                  NNNNEE..    NNNNNN..    NNNNEE..
                BC                  NNNNNN..    NNNNEE..    NNNNEE..
                                    ^^^^        ^^^^        ^^^^
                Range               C           A           AC
                Tag                 N*          N*          N
            Please note the high end of range is unavailable if the continuing
            side is "N" and not complete deletion. (marked "N*")
            All "N" and "N*" containing ranges are conflict but are checked for
            A==C check (case "e")

            Here tag for diff3:
                'E'    a[j1:j2] == b[i1:i2] == c[k1:k2] == a[j1:j2]
                'A'    a[j1:j2] != b[i1:i2] == c[k1:k2] != a[j1:j2]
                'C'    a[j1:j2] == b[i1:i2] != c[k1:k2] != a[j1:j2]
                'e'    a[j1:j2] != b[i1:i2] != c[k1:k2] == a[j1:j2]
                'N'    a[j1:j2] != b[i1:i2] != c[k1:k2] != a[j1:j2]

    This requires diff2lib which is a wrapper of difflib and provides
    'get_opcodes' method. Here tag for diff2:

                'E'    a[j1:j2] == b[i1:i2]
                'N'    a[j1:j2] != b[i1:i2]

    Example1: comparing two strings, and considering None to be "junk"

    >>> a = "private Thread currentThread foo bar;"
    >>> b = "private volatile Thread currentThread;"
    >>> c = "private volatile currentThread foo;tail"
    >>> s = SequenceMatcher3(a, b, c)
    >>>

    If you want to know how to change the first sequence into the second,
    use .get_opcodes():

    >>> for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
    ...     print("%s a[%d:%d],b[%d:%d],c[%d:%d] => '%s', '%s', '%s'" % (
    ...     tag, i1, i2, j1, j2, k1, k2, a[i1:i2], b[j1:j2], c[k1:k2]))
    ...
    E a[0:6],b[0:6],c[0:6] => 'privat', 'privat', 'privat'
    A a[6:6],b[6:15],c[6:15] => '', 'e volatil', 'e volatil'
    E a[6:8],b[15:17],c[15:17] => 'e ', 'e ', 'e '
    C a[8:15],b[17:24],c[17:17] => 'Thread ', 'Thread ', ''
    E a[15:28],b[24:37],c[17:30] => 'currentThread', 'currentThread', 'currentThread'
    N a[28:36],b[37:37],c[30:34] => ' foo bar', '', ' foo'
    E a[36:37],b[37:38],c[34:35] => ';', ';', ';'
    C a[37:37],b[38:38],c[35:39] => '', '', 'tail'

    Example 2: comparing two strings, and considering blanks to be "junk"

    >>> a = "private Thread currentThread foo bar;"
    >>> b = "private volatile Thread currentThread;"
    >>> c = "private volatile currentThread foo;tail"
    >>> s = SequenceMatcher3(a, b, c, 0, lambda x: x == " ", True)
    >>>

    If you want to know how to change the first sequence into the second,
    use .get_opcodes():

    >>> for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
    ...     print("%s a[%d:%d],b[%d:%d],c[%d:%d] => '%s', '%s', '%s'" % (
    ...     tag, i1, i2, j1, j2, k1, k2, a[i1:i2], b[j1:j2], c[k1:k2]))
    ...
    E a[0:8],b[0:8],c[0:8] => 'private ', 'private ', 'private '
    A a[8:8],b[8:16],c[8:16] => '', 'volatile', 'volatile'
    e a[8:8],b[16:17],c[16:16] => '', ' ', ''
    C a[8:14],b[17:23],c[16:16] => 'Thread', 'Thread', ''
    E a[14:28],b[23:37],c[16:30] => ' currentThread', ' currentThread', ' currentThread'
    N a[28:36],b[37:37],c[30:34] => ' foo bar', '', ' foo'
    E a[36:37],b[37:38],c[34:35] => ';', ';', ';'
    C a[37:37],b[38:38],c[35:39] => '', '', 'tail'

    This example 2 produces more intuitive result due to setting for "junk".

    Methods:

    __init__(a='', b='', c='', matcher=0, isjunk=None, autojunk=True, linerule=2)
        Construct a SequenceMatcher3.

    set_seqs(a, b)
        Set the two sequences to be compared.

    set_seq1(a)
        Set the first sequence to be compared.

    set_seq2(b)
        Set the second sequence to be compared.

    set_seq3(c)
        Set the third sequence to be compared.

    get_opcodes()
        Return list of 7-tuples describing how to merge c into a while b being
        common older version.
    """

    def __init__(
        self, a="", b="", c="", matcher=0, isjunk=None, autojunk=True, linerule=2
    ):
        """Construct a SequenceMatcher3.

        Optional arg a is the first of three sequences to be compared.  By
        default, an empty string.  The elements of a must be hashable.  See
        also .set_seqs() and .set_seq1().

        Optional arg b is the second of three sequences to be compared.  By
        default, an empty string.  The elements of b must be hashable. See
        also .set_seqs() and .set_seq2().

        Optional arg c is the third of three sequences to be compared.  By
        default, an empty string.  The elements of b must be hashable. See
        also .set_seqs() and .set_seq3().

        Optional arg matcher should be set to 0 for hashables such as strings,
        or 1 for lists of lines.

        Optional arg isjunk is None (the default), or a one-argument
        function that takes a sequence element and returns true iff the
        element is junk.  None is equivalent to passing "lambda x: 0", i.e.
        no elements are considered to be junk.  For example, pass
            lambda x: x in " \\t"
        if you're comparing lines as sequences of characters, and don't
        want to synch up on blanks or hard tabs.

        Optional arg autojunk should be set to False to disable the
        "automatic junk heuristic" that treats popular elements as junk
        (see module documentation for more information).
        """

        # Members:
        # a
        #      first sequence (yours, merge prioritize on this)
        # b
        #      second sequence (old common, used to check who changed what)
        # c
        #      third sequence (theirs, data to be merged)
        # opcodes
        #      a list of (tag, i1, i2, j1, j2, k1, k2) tuples, where tag is
        #      one of
        #          'E'    a[i1:i2] == b[j1:j2] == c[k1:k2] == a[i1:i2]
        #          'A'    a[i1:i2] != b[j1:j2] == c[k1:k2] != a[i1:i2]
        #          'C'    a[i1:i2] == b[j1:j2] != c[k1:k2] != a[i1:i2]
        #          'e'    a[i1:i2] != b[j1:j2] != c[k1:k2] == a[i1:i2]
        #          'N'    a[i1:i2] != b[j1:j2] != c[k1:k2] != a[i1:i2]
        # isjunk
        #      a user-supplied function taking a sequence element and
        #      returning true iff the element is "junk" -- this has
        #      subtle but helpful effects on the algorithm, which I'll
        #      get around to writing up someday <0.9 wink>.
        #      DON'T USE!  Only __chain_b uses this.  Use "in self.bjunk".

        self.a = a
        self.b = b
        self.c = c
        self.matcher = matcher
        self.isjunk = isjunk
        self.autojunk = autojunk
        self.linerule = linerule
        self.opcodes = None

    def set_seq1(self, a):
        """Set the first sequence to be compared."""

        if a is self.a:
            return
        self.a = a
        self.opcodes = None

    def set_seq2(self, b):
        """Set the second sequence to be compared."""

        if b is self.b:
            return
        self.b = b
        self.opcodes = None

    def set_seq3(self, c):
        """Set the third sequence to be compared."""

        if c is self.c:
            return
        self.c = c
        self.opcodes = None

    def set_seqs(self, a, b, c):
        """Set the two sequences to be compared."""

        self.set_seq1(a)
        self.set_seq2(b)
        self.set_seq3(c)

    def get_opcodes(self):
        """Return list of 7-tuples describing how a, b, c matches.

        Each tuple is of the form (tag, i1, i2, j1, j2, k1, k2).  The first
        tuple has i1 == j1 == k1 == 0, and remaining tuples have i1 == the i2
        from the tuple preceding it, and likewise for j1 == the previous j2,
        and likewise for k1 == the previous k2.

        The tags are strings, with these meanings:

        'E':   a == b == c
        'e':   a != b != c == A
        'A':   c == b --> a, change in a
        'C':   a == b --> c, change in c
        'N':   b-->a, b-->c, conflicting changes

        >>> a = "qabxcdsdgp"
        >>> b = "abycdfsdkg"
        >>> c = "abycdfzcpgp"
        >>> s = SequenceMatcher3(a, b, c)
        >>> for tag, i1, i2, j1, j2, k1, k2 in s.get_opcodes():
        ...    print(("%s a[%d:%d] (%s) / b[%d:%d] (%s) / c[%d:%d] (%s)" %
        ...           (tag, i1, i2, a[i1:i2], j1, j2, b[j1:j2], k1, k2, c[k1:k2])))
        ...
        A a[0:1] (q) / b[0:0] () / c[0:0] ()
        E a[1:3] (ab) / b[0:2] (ab) / c[0:2] (ab)
        A a[3:4] (x) / b[2:3] (y) / c[2:3] (y)
        E a[4:6] (cd) / b[3:5] (cd) / c[3:5] (cd)
        A a[6:6] () / b[5:6] (f) / c[5:6] (f)
        N a[6:8] (sd) / b[6:9] (sdk) / c[6:9] (zcp)
        E a[8:9] (g) / b[9:10] (g) / c[9:10] (g)
        e a[9:10] (p) / b[10:10] () / c[10:11] (p)
        """

        a = self.a
        b = self.b
        c = self.c
        matcher = self.matcher
        if matcher == 0:
            opcodes_ba = SequenceMatcher(self.isjunk, b, a).get_opcodes()
            opcodes_bc = SequenceMatcher(self.isjunk, b, c).get_opcodes()
            tag_equal = "equal"
        else:  # matcher == 1
            opcodes_ba = LineMatcher(b, a, self.linerule).get_opcodes()
            opcodes_bc = LineMatcher(b, c, self.linerule).get_opcodes()
            tag_equal = "E"
        n_ba = 0  # walking index for opcodes_ba
        n_bc = 0  # walking index for opcodes_bc
        len_ba = len(opcodes_ba)
        len_bc = len(opcodes_bc)
        logger.debug(
            "diff3lib: matcher={} tag={} len_ba={} len_bc={}".format(
                matcher, tag_equal, len_ba, len_bc
            )
        )
        il = jl = kl = 0  # range lower end for b, a, c (next in next round)
        ih = jh = kh = 0  # range high end for b, a, c (next in next round)
        answer = list()
        tag = ""
        while n_ba < len_ba or n_bc < len_bc:
            # get a chunk data
            if n_ba < len_ba:
                (tag_ba, il_ba, ih_ba, jl_ba, jh_ba) = opcodes_ba[n_ba]
                logger.debug(
                    "diff3lib: NORMAL n_bc={} < len_bc={}".format(n_bc, len_bc)
                )
            elif len_ba == 0:
                (tag_ba, il_ba, ih_ba, jl_ba, jh_ba) = (tag_equal, 0, 0, 0, 0)
                logger.debug(
                    "diff3lib: UNDERRUN n_ba={} len_ba={}".format(n_ba, len_ba)
                )
            else:
                (tag_ba, il_ba, ih_ba, jl_ba, jh_ba) = opcodes_ba[len_ba - 1]
                (tag_ba, il_ba, ih_ba, jl_ba, jh_ba) = (
                    tag_equal,
                    ih_ba,
                    ih_ba,
                    jh_ba,
                    jh_ba,
                )
                logger.debug("diff3lib: OVERRUN n_ba={} len_ba={}".format(n_ba, len_ba))
            if n_bc < len_bc:
                (tag_bc, il_bc, ih_bc, kl_bc, kh_bc) = opcodes_bc[n_bc]
                logger.debug(
                    "diff3lib: NORMAL n_bc={} < len_bc={}".format(n_bc, len_bc)
                )
            elif len_bc == 0:
                (tag_bc, il_bc, ih_bc, kl_bc, kh_bc) = (tag_equal, 0, 0, 0, 0)
                logger.debug(
                    "diff3lib: UNDERRUN n_bc={} len_bc={}".format(n_bc, len_bc)
                )
            else:
                (tag_bc, il_bc, ih_bc, kl_bc, kh_bc) = opcodes_bc[len_bc - 1]
                (tag_bc, il_bc, ih_bc, kl_bc, kh_bc) = (
                    tag_equal,
                    ih_bc,
                    ih_bc,
                    kh_bc,
                    kh_bc,
                )
                logger.debug("diff3lib: OVERRUN n_bc={} len_bc={}".format(n_bc, len_bc))
            # get tag for this set of opcodes if high range value is available
            if tag == "N":
                pass  # All undecided comes in as tag == "N", otherwise tag == ""
            elif tag_ba == tag_equal and tag_bc == tag_equal:
                tag = "E"
            elif tag_ba != tag_equal and tag_bc == tag_equal:
                tag = "A"
            elif tag_ba == tag_equal and tag_bc != tag_equal:
                tag = "C"
            elif tag_ba != tag_equal and tag_bc != tag_equal:
                tag = "N"
            if ih_ba == ih_bc:
                n_ba += 1
                n_bc += 1
                ih = ih_ba  # == ih_bc
                jh = jh_ba
                kh = kh_bc
            elif ih_ba > ih_bc:
                n_bc += 1
                ih = ih_bc
                kh = kh_bc
                if tag_ba == tag_equal:
                    jh = jh_ba - (ih_ba - ih_bc)
                elif jh_ba == jl_ba:
                    jh = jh_ba
                else:
                    jh = None  # undecided
                    tag = "N"
            if ih_ba == ih_bc:
                kh = kh_bc
            elif ih_ba < ih_bc:
                n_ba += 1
                ih = ih_ba
                jh = jh_ba
                if tag_bc == tag_equal:
                    kh = kh_bc - (ih_bc - ih_ba)
                elif kh_bc == kl_bc:
                    kh = kh_bc
                else:
                    kh = None  # undecided
                    tag = "N"
            if jh is None:
                logger.debug(
                    "diff3lib: UNDECIDED_ba tag={}, jl={}, jh=None, il={}, ih={}, kl={}, kh={}".format(
                        tag, jl, il, ih, kl, kh
                    )
                )
            elif kh is None:
                logger.debug(
                    "diff3lib: UNDECIDED_bc tag={}, jl={}, jh={}, il={}, ih={}, kl={}, kh=None".format(
                        tag, jl, jh, il, ih, kl
                    )
                )
            else:
                if tag == "N":
                    if a[jl:jh] == c[kl:kh]:
                        tag = "e"
                answer.append((tag, jl, jh, il, ih, kl, kh))
                logger.debug(
                    "diff3lib: APPEND tag={}, jl={}, jh={}, il={}, ih={}, kl={}, kh={}".format(
                        tag, jl, jh, il, ih, kl, kh
                    )
                )
                il = ih
                jl = jh
                kl = kh
                tag = ""
        self.opcodes = answer
        return answer


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/python3
# vi:se ts=4 sts=4 et ai:
"""
Copyright (C) 2018-2021       Osamu Aoki <osamu@debian.org>

This is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2,
or (at your option) any later version.

This is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with the program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""
import setuptools

setuptools.setup()



#!/usr/bin/python3
# vim:se tw=0 sts=4 ts=4 et ai:
"""
Script to test imediff macro key translation
"""
import imediff

args = imediff.initialize_args()
confs = imediff.initialize_confs(args.conf)
if args.diff_mode == 0 or args.diff_mode is None:
    args.diff_mode = 2
list_a = ["a", "b", "c", "d", "e", "f"]
list_b = ["a", "b", "c", "d", "e", "f"]
list_c = ["a", "b", "c", "d", "e", "f"]
macro = "".join([chr(i) for i in range(ord("a"), ord("z"))])
macro += "".join([chr(i) for i in range(ord("A"), ord("Z"))])
args.macro = macro  # override
args.edit_cmd = "DUMMY_EDITOR"
instance = imediff.TextData(list_a, list_b, list_c, args, confs)
for c in macro:
    k = instance.getch_translated()
    print("MACRO:{} --> TranslatedKey={}".format(c, chr(k)))
